Set Implicit Arguments.

Require Import Lists.List.

Inductive Number : Set := Num : nat -> Number.
Inductive String : Set := Str : nat -> String.

Axiom Dec_String : forall (x x' : String),  x = x' \/ x <> x'.
Axiom Dec_Number : forall (x x' : Number),  x = x' \/ x <> x'.


(* ******************************************************************************** *)
(* Maps *)

Section Maps.

Variable A B : Type.

Axiom Update : (A -> B) -> A -> B -> (A -> B).
Axiom Singleton : A -> B -> (A -> B).

Axiom Fresh : (A -> B) -> A -> Prop.

Axiom Update_Same : 
  forall m a b, 
  (Update m a b) a = b .

Axiom Update_Different : 
  forall m a b a', 
    a <> a' ->
  (Update m a b) a' = m a'.

Axiom Fresh_Update :
  forall m a b a',
    Fresh m a ->
  (Update m a b) a' = m a'.

End Maps.

Axiom Singleton_Option_Same : 
   forall A B (a : A) (b : B),
     (Singleton a (Some b)) a = Some b.

Axiom Singleton_Option_Different : 
   forall A B (a : A) (b : B) (a' : A),
     a <> a' ->
     (Singleton a (Some b)) a' = None.

(* ******************************************************************************** *)
(* Bijections *)

Section Bijections.

(* Low-equivalence *)

(* Bijections *)
Require Import Relations.

Definition Injective (A : Type) (R : relation A) := forall x y z, R x z -> R y z -> x = y.

Record Bijection (A : Type) : Type :=
{ 
  extension : relation A;
  bijective : Injective extension /\ Injective (transp _ extension);
}.

Axiom Bijection_Prop : 
  forall A (a1 a1' a2 a2' : A) (b : Bijection A), 
    (extension b) a1 a2 ->
    (extension b) a1' a2' ->
    (a1 <> a1' /\ a2 <> a2') \/ (a1 = a1' /\ a2 = a2'). 

End Bijections.

