Set Implicit Arguments.

Require Import Basics.
Require Import Syntax.
Require Import Semantics.
Require Import LowEq.
Require Import ModelHypotheses.

(* ******************************************************************************** *)
(* If fresh, then not in bijections, and then can be added w/o changing low-equivalence *)

Inductive Fresh_MStack : list MPointer -> MPointer -> Prop :=
  | frMNil : forall p, Fresh_MStack nil p
  | frMCons : forall p' p ps, p' <> p -> Fresh_MStack ps p' -> Fresh_MStack (p :: ps) p'. 

(* Requires reasoning that p is not in bijection *)
Axiom Conf_Fresh_MUpdate :
  forall lρ uρ mρ p lσ uσ mσ x, 
    LowEq_Memory (MkStack lρ uρ mρ, MkHeap lσ uσ mσ) (MkStack lρ uρ mρ, MkHeap lσ uσ mσ) ->
    Fresh_MStack mρ p ->
  LowEq_Memory (MkStack lρ uρ mρ, MkHeap lσ uσ mσ) (MkStack lρ uρ mρ, MkHeap lσ uσ (Update mσ p x)).


Axiom Fresh_MPointer_Fresh_MStack : 
   forall lρ uρ mρ lσ uσ mσ p, 
     LowEq_Memory (MkStack lρ uρ mρ, MkHeap lσ uσ mσ) (MkStack lρ uρ mρ, MkHeap lσ uσ mσ) ->
     Fresh mσ p ->
   Fresh_MStack mρ p.


(* ******************************************************************************** *)

Inductive Fresh_LStack : list LPointer -> LPointer -> Prop :=
  | frLNil : forall p, Fresh_LStack nil p
  | frLCons : forall p' p ps, p' <> p -> Fresh_LStack ps p' -> Fresh_LStack (p :: ps) p'. 

Axiom Conf_Fresh_LUpdate : 
  forall lρ uρ mρ σl σu σm p v, 
    LowEq_Memory (MkStack lρ uρ mρ, MkHeap σl σu σm) (MkStack lρ uρ mρ, MkHeap σl σu σm) ->
    Fresh_LStack lρ p ->
  LowEq_Memory (MkStack lρ uρ mρ, MkHeap σl σu σm) (MkStack lρ uρ mρ, MkHeap (Update σl p (LIValue v)) σu σm).

Axiom Fresh_LPointer_Fresh_LStack : 
   forall lρ uρ mρ lσ uσ mσ p, 
     LowEq_Memory (MkStack lρ uρ mρ, MkHeap lσ uσ mσ) (MkStack lρ uρ mρ, MkHeap lσ uσ mσ) ->
     Fresh lσ p ->
   Fresh_LStack lρ p.

(* ******************************************************************************** *)

Lemma Conf_FindM : 
  forall lρ uρ mρ lσ uσ mσ α x ,
    LowEq_Memory (MkStack lρ uρ mρ, MkHeap lσ uσ mσ) (MkStack lρ uρ mρ , MkHeap lσ uσ mσ) ->
    FindM mρ mσ α x -> 
  (exists p, x = Some p /\ extension bm p p) \/ (x = None).
Proof.
intros lρ uρ mρ lσ uσ mσ α x LE_M1M1 FH. 
inversion LE_M1M1 as [ Γ1 Γ2 Σ1 Σ2 LE_LSLS LE_LHLH LE_MSMS LE_MHMH]. subst. simpl in *.
induction mρ as [ | p ps].

(* nil *)
inversion FH.
auto.

(* p : ps *)
inversion FH as [ _mσ _x | _mσ _x _p _ps ω v M Fr | _mσ _x _p _ps ω v M Fr Fi ]; subst.

(* found in p *)
inversion LE_MSMS.
left. exists p; auto.

(* found in ps *)
inversion LE_MSMS.
apply IHps; auto.
constructor; auto.
Qed.

(* ******************************************************************************** *)

Lemma Conf_UpdateM : 
  forall Γ lσ uσ mσ mσ' α l, 
    LowEq_Memory (Γ, MkHeap lσ uσ mσ) (Γ, MkHeap lσ uσ mσ) ->
    UpdateM H (mstack Γ) mσ α l mσ' ->
  LowEq_Memory (Γ, MkHeap lσ uσ mσ) (Γ, MkHeap lσ uσ mσ').
Proof.
intros Γ lσ uσ mσ mσ' α l LE_M1M1 U.
inversion LE_M1M1 as [_Γ _Γ' _Σ _Σ' LE_LSLS LE_LHLH LE_MSMS LE_MHMH]. subst.
constructor; auto.

inversion U as [ p ps _mσ1 _mσ2 _α _l ω1 ω2 FM HH UF UH | ctx _mρ p _mσ1 _mσ2 _α l1 l2 ω1 ω2 FM HH FH CF UF UH]; subst.

destruct Γ as [lρ uρ m].
simpl in *.

destruct (Conf_FindM LE_M1M1 FM) as [ [_p [EQ bij1]] | EQ]; try congruence.
assert (_p = p). congruence. subst. clear EQ.

assert (LE_MIM1 :  LowEq_MItem (mσ p) (mσ p)).
apply LE_MHMH; auto.

intros p1 p2 bij2.
assert (LE_MIM2 :  LowEq_MItem (mσ p1) (mσ p2)).
apply LE_MHMH; auto.

destruct (Bijection_Prop _ _ _ _ _ bij1 bij2) as [[EQ1 EQ2] | [EQ1 EQ2]].
rewrite (Update_Different _ _ EQ2).
assumption.

subst.
rewrite Update_Same.
rewrite HH.
constructor.
constructor.

intro x. 
destruct (Dec_String α x) as [ EQ | EQ ].
subst. inversion CF. subst. rewrite Lub_Secret.
left. exists (MVLabel H), (MVLabel H).
split. assumption.
split. rewrite Update_Same. trivial.
constructor.

rewrite (Update_Different _ _ EQ).
inversion LE_MIM1 as [_ω1 _ω1' LE_FrFr]. subst.
inversion LE_FrFr as [_ω2 _ω2' LE_FrFr_Prop]. subst.
assert (_ω1 = ω1). congruence. subst.
apply LE_FrFr_Prop.
Qed.

(* ******************************************************************************** *)

Lemma Conf_DefineM : 
  forall lρ uρ mρ p lσ uσ mσ mσ' α l, 
    LowEq_Memory (MkStack lρ uρ mρ, MkHeap lσ uσ mσ) (MkStack lρ uρ mρ, MkHeap lσ uσ mσ) ->
    Fresh_MStack mρ p ->
    DefineM (p :: mρ) mσ α l mσ' ->
  LowEq_Memory (MkStack lρ uρ mρ, MkHeap lσ uσ mσ) (MkStack lρ uρ mρ, MkHeap lσ uσ mσ').
Proof.
intros lρ uρ mρ p lσ uσ mσ mσ' α l LE_M1M1 FrH D.
inversion LE_M1M1 as [_Γ _Γ' _Σ _Σ' LE_LSLS LE_LHLH LE_MSMS LE_MHMH]. subst.
inversion D; subst.
apply Conf_Fresh_MUpdate; auto.
Qed.

Lemma Conf_UpdateM' : 
  forall lρ uρ mρ p lσ uσ mσ mσ' α l, 
    LowEq_Memory (MkStack lρ uρ mρ, MkHeap lσ uσ mσ) (MkStack lρ uρ mρ, MkHeap lσ uσ mσ) ->
    Fresh_MStack mρ p ->
    UpdateM H (p :: mρ) mσ α l mσ' ->
  LowEq_Memory (MkStack lρ uρ mρ, MkHeap lσ uσ mσ) (MkStack lρ uρ mρ, MkHeap lσ uσ mσ').
Proof.
intros lρ uρ mρ p' lσ uσ mσ mσ' α l LE_M1M1 FrH U.
inversion U as [ | ctx _mρ p _mσ1 _mσ2 _α l1 l2 ω1 ω2 FM HH FH CF UF _UH]; subst.

inversion FM as [ | _mσ x θ θs ω v HH' FH' | _mσ x _p' ps ω v HH' FH' FM']. subst.
(* Found in first *)
apply Conf_Fresh_MUpdate; auto.

(* Not found in first *)
subst.
inversion LE_M1M1 as [_Γ _Γ' _Σ _Σ' LE_LSLS LE_LHLH LE_MSMS LE_MHMH]. subst.
constructor; auto.
simpl.

destruct (Conf_FindM LE_M1M1 FM') as [ [_p [EQ bij1]] | EQ]; try congruence.
assert (_p = p). congruence. subst. clear EQ.

assert (LE_MIM1 :  LowEq_MItem (mσ p) (mσ p)).
apply LE_MHMH; auto.

intros p1 p2 bij2.
assert (LE_MIM2 :  LowEq_MItem (mσ p1) (mσ p2)).
apply LE_MHMH; auto.

destruct (Bijection_Prop _ _ _ _ _ bij1 bij2) as [[EQ1 EQ2] | [EQ1 EQ2]].
rewrite (Update_Different _ _ EQ2).
assumption.

subst.
rewrite Update_Same.
rewrite HH.
constructor.
constructor.

intro x. 
destruct (Dec_String α x) as [ EQ | EQ ].
subst. inversion CF. subst. rewrite Lub_Secret.
left. exists (MVLabel H), (MVLabel H).
split. assumption.
split. rewrite Update_Same. trivial.
constructor.

rewrite (Update_Different _ _ EQ).
inversion LE_MIM1 as [_ω1 _ω1' LE_FrFr]. subst.
inversion LE_FrFr as [_ω2 _ω2' LE_FrFr_Prop]. subst.
assert (_ω1 = ω1). congruence. subst.
apply LE_FrFr_Prop.
Qed.

(* ******************************************************************************** *)

Lemma Conf_Interp_EffC : 
  forall lρ uρ mρ p lσ uσ mσ Σ2 c,
     LowEq_Memory (MkStack lρ uρ mρ, MkHeap lσ uσ mσ) (MkStack lρ uρ mρ, MkHeap lσ uσ mσ) -> 
     Fresh_MStack mρ p ->
     Interp_EffC H (MkStack lρ uρ (p :: mρ)) (MkHeap lσ uσ mσ) c Σ2 ->
  LowEq_Memory (MkStack lρ uρ mρ, MkHeap lσ uσ mσ) (MkStack lρ uρ mρ, Σ2).
Proof.
intros lρ uρ mρ p lσ uσ mσ Σ2 c LE_M1M1 FrH IH.
inversion LE_M1M1 as [_Γ _Γ' _Σ _Σ' LE_LSLS LE_LHLH LE_MSMS LE_MHMH]. subst.

induction c.

(* ECRefRead *)
inversion IH. subst. 
eapply Conf_DefineM; eauto.

(* ECRefWrite *)
inversion IH. subst. 
eapply Conf_DefineM; eauto.

(* ECApp *)
inversion IH. subst. 
eapply Conf_DefineM; eauto.

(* ECAsn *)
inversion IH. subst.
eapply Conf_UpdateM'; eauto.
Qed.

(* ******************************************************************************** *)

Lemma Conf_Interp_EffCs : 
  forall ξ lρ uρ mρ p, 
     Fresh_MStack mρ p ->
     forall Σ1 Σ2,
     LowEq_Memory (MkStack lρ uρ mρ, Σ1) (MkStack lρ uρ mρ, Σ1) -> 
     Interp_EffCs H (MkStack lρ uρ (p :: mρ)) Σ1 ξ Σ2 ->
  LowEq_Memory (MkStack lρ uρ mρ, Σ1) (MkStack lρ uρ mρ, Σ2).
Proof.
intros ξ lρ uρ mρ p FrH.
induction ξ. 
(* nil *)
intros Σ1 Σ2 LE_M1M1 IH.
inversion IH. subst.
assumption.
(* cons *)
intros Σ1 Σ2 LE_M1M1 IcH.
rename Σ2 into Σ3.
inversion IcH as [ | ctx Γ _Σ1 Σ2 _Σ3 c cs IH IcH_Sub ]. subst.
assert (LE_M1M2 : LowEq_Memory (MkStack lρ uρ mρ, Σ1) (MkStack lρ uρ mρ, Σ2)).
destruct Σ1 as [lρ1 uρ1 mρ1]. destruct Σ2 as [lρ2 uρ2 mρ2].
eapply Conf_Interp_EffC; eauto.

assert (LE_M2M2 : LowEq_Memory (MkStack lρ uρ mρ, Σ2) (MkStack lρ uρ mρ, Σ2)).
eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (MkStack lρ uρ mρ, Σ2) (MkStack lρ uρ mρ, Σ3)).
destruct Σ2 as [lρ2 uρ2 mρ2].
eapply IHξ; eauto.
eapply LowEq_Memory_Trans.
  apply LE_M1M2.
  apply LE_M2M3.
Qed.

(* ******************************************************************************** *)

Lemma Conf_MkLRef :
  forall Γ Σ1 Σ2 v LR LW,
    MkLRef Σ1 v (Σ2, LR, LW) -> 
    LowEq_Memory (Γ, Σ1) (Γ, Σ1) ->  
  LowEq_Memory (Γ, Σ1) (Γ, Σ2).
Proof.
intros Γ Σ1 Σ2 v LR LW MK LE_M1M1.
inversion MK.
subst.
destruct Γ.
apply Conf_Fresh_LUpdate.
  apply LE_M1M1.
  eapply Fresh_LPointer_Fresh_LStack; eauto.
Qed.

Lemma Conf_MkURef :
  forall Γ Σ1 Σ2 v UR UW,
    MkURef Σ1 v (Σ2, UR, UW) -> 
    LowEq_Memory (Γ, Σ1) (Γ, Σ1) ->  
  LowEq_Memory (Γ, Σ1) (Γ, Σ2).
Proof.
intros Γ Σ1 Σ2 v LR LW MK LE_M1M1.
inversion MK. subst.
eapply LowEq_Memory_UHeap_ModR.
apply LE_M1M1.
Qed.

(* ******************************************************************************** *)

Lemma Conf_Unlabel_LValue :
  forall ctx Γ Σ1 Σ2 lv1 ϕ uv1, 
    Unlabel_LValue ctx  Γ Σ1 lv1 ϕ uv1 Σ2 -> 
    ctx = H ->
    LowEq_Memory (Γ, Σ1) (Γ, Σ1) -> 
  LowEq_Memory (Γ, Σ1) (Γ, Σ2).
Proof.
intros ctx Γ Σ1 Σ2 lv1 ϕ uv1 UL.
induction UL as [ ctx Γ Σ1 Σ2 mσ l n α U Mk 
                | ctx Γ Σ1 Σ2 mσ l LF α π U Mk 
                | ctx Γ Σ1 Σ2 mσ l LR LW α π U Mk
                | ctx Γ Σ1 Σ2 mσ l α ϕ U Mk 
                | ctx Γ Σ1 Σ2 mσ l Hd Tl α ϕ U Mk ].

(* Prim *)
intros c LE_MM.
subst.

destruct Σ1.
eapply Conf_UpdateM.
  apply LE_MM.
  apply U.
 
(* Fun *)
intros c LE_MM.
subst.

destruct Σ1.
eapply Conf_UpdateM.
  apply LE_MM.
  apply U.

(* Ref *)
intros c LE_MM.
subst.

destruct Σ1.
eapply Conf_UpdateM.
  apply LE_MM.
  apply U.

(* Nil *)
intros c LE_MM.
subst.

destruct Σ1.
eapply Conf_UpdateM.
  apply LE_MM.
  apply U.

(* Cons *)
intros c LE_MM.
subst.

destruct Σ1.
eapply Conf_UpdateM.
  apply LE_MM.
  apply U.
Qed.



(* ******************************************************************************** *)

Definition Conf_LRefReadEval_Prop LR ctx Γ Σ1 Σ2 lv (EV : LRefReadEval LR ctx Γ Σ1 Σ2 lv) :=
  ctx = H -> 
  LowEq_Memory (Γ, Σ1) (Γ, Σ1) -> 
  LowEq_Memory (Γ, Σ1) (Γ, Σ2).

Definition Conf_URefReadEval_Prop UR ctx Γ Σ1 Σ2 uv (EV : URefReadEval UR ctx Γ Σ1 Σ2 uv) :=
  ctx = H -> 
  LowEq_Memory (Γ, Σ1) (Γ, Σ1) -> 
  LowEq_Memory (Γ, Σ1) (Γ, Σ2).

Lemma Conf_RefReadEval : 
  (forall LR ctx Γ Σ1 Σ2 lv (EV : LRefReadEval LR ctx Γ Σ1 Σ2 lv), Conf_LRefReadEval_Prop EV) /\   
  (forall UR ctx Γ Σ1 Σ2 lv (EV : URefReadEval UR ctx Γ Σ1 Σ2 lv), Conf_URefReadEval_Prop EV).
Proof.
apply RefReadEval_mutind with 
  (P  := Conf_LRefReadEval_Prop)
  (P0 := Conf_URefReadEval_Prop); intros.

(* LLRefRead *)
unfold Conf_LRefReadEval_Prop.
intros _t LE_M1M1.

apply LE_M1M1.

(* LURefRead *)
rename H into IHEV.
 
unfold Conf_LRefReadEval_Prop.
intros _t LE_M1M1.
subst ctx.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV.
  trivial.
  apply LE_M1M1.

apply LE_M1M2.

(* UURefRead *)
unfold Conf_LRefReadEval_Prop.
intros _t LE_M1M1.

apply LE_M1M1.

(* ULRefRead *)
rename H into IHEV.

unfold Conf_LRefReadEval_Prop.
intros _t LE_M1M1.
subst ctx.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV.
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply Conf_Unlabel_LValue.
  apply u.
  trivial.
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
apply LE_M2M3.
Qed.

Lemma Conf_LRefReadEval : 
  forall LR ctx Γ Σ1 Σ2 lv (EV : LRefReadEval LR ctx Γ Σ1 Σ2 lv), Conf_LRefReadEval_Prop EV.
Proof.
apply Conf_RefReadEval.
Qed.

Lemma Conf_URefReadEval : 
  forall UR ctx Γ Σ1 Σ2 lv (EV : URefReadEval UR ctx Γ Σ1 Σ2 lv), Conf_URefReadEval_Prop EV.
Proof.
apply Conf_RefReadEval.
Qed.


Lemma Conf_Update_Secret : 
  forall Γ lσ uσ mσ v1 v2 p, 
     LowEq_Memory (Γ, MkHeap lσ uσ mσ) (Γ, MkHeap lσ uσ mσ) ->
     lσ p = LIValue v1 ->
     CanFlow H (LabelOf v1) ->
  LowEq_Memory 
    (Γ, MkHeap lσ uσ mσ) 
    (Γ, MkHeap (Update lσ p (LIValue (Upgrade H v2))) uσ mσ).
Proof.
intros Γ lσ uσ mσ v1 v2 p LE_MM l c.

inversion LE_MM as [_Γ _Γ' _Σ _Σ' LE_LSLS LE_LHLH LE_MSMS LE_MHMH]. subst.
constructor; try assumption.

intros p' p'' bij.

destruct (Dec_LPointer p p'') as [EQ | EQ].
(* p = p'' *)
subst.
assert (LE_p'' : LowEq_LItem (lσ p') (lσ p'')).
apply LE_LHLH.
  apply bij.
simpl. rewrite (Update_Same lσ p'').

rewrite l in LE_p''.
inversion LE_p'' as [ v1' _v2 LE_v1v2 | ]. subst.
inversion LE_v1v2; subst; inversion c. 
constructor.
constructor.
  assumption.
  apply LabelOf_Upgrade_Secret.
(* p <> p'' *)
simpl. 
rewrite (Update_Different lσ (LIValue (Upgrade H v2)) EQ).
apply LE_LHLH.
  apply bij.
Qed.


(* ******************************************************************************** *)

Definition Conf_LRefWriteEval_Prop LW ctx Γ Σ1 lv Σ2 (EV : LRefWriteEval LW ctx Γ Σ1 lv Σ2) :=
  ctx = H -> LowEq_Memory (Γ, Σ1) (Γ, Σ1) -> LowEq_Memory (Γ, Σ1) (Γ, Σ2).

Definition Conf_URefWriteEval_Prop UW ctx Γ Σ1 lv Σ2 (EV : URefWriteEval UW ctx Γ Σ1 lv Σ2) :=
  ctx = H -> LowEq_Memory (Γ, Σ1) (Γ, Σ1)  -> LowEq_Memory (Γ, Σ1) (Γ, Σ2).

Lemma Conf_RefWriteEval : 
  (forall LW ctx Γ Σ1 lv Σ2 (EV : LRefWriteEval LW ctx Γ Σ1 lv Σ2), Conf_LRefWriteEval_Prop EV) /\
  (forall UW ctx Γ Σ1 lv Σ2 (EV : URefWriteEval UW ctx Γ Σ1 lv Σ2), Conf_URefWriteEval_Prop EV).
Proof.
apply RefWriteEval_mutind with 
  (P  := Conf_LRefWriteEval_Prop)
  (P0 := Conf_URefWriteEval_Prop); intros.

(* LLRefWrite *)
unfold Conf_LRefWriteEval_Prop.
intros _t LE_M1M1.
subst ctx. 

assert (LE_M1M2: LowEq_Memory (Γ, Σ1) (Γ, Σ2)). subst.
destruct Σ1. 
eapply Conf_Update_Secret.
  apply LE_M1M1.
  apply e.
  apply c.

apply LE_M1M2. 

(* LURefWrite *)
rename H into IHEV. 

unfold Conf_LRefWriteEval_Prop.
intros _t LE_M1M1.
subst ctx.

assert (LE_M1M1_cl : 
  LowEq_Memory 
    (MkStack (lstack Γ) (ustack Γ) mρ, Σ1) 
    (MkStack (lstack Γ) (ustack Γ) mρ, Σ1)
).
apply LowEq_MClosure_Intro.
  apply LowEq_MStack_Refl.
  apply LE_M1M1.

assert (LE_M1M2_cl: 
  LowEq_Memory 
    (MkStack (lstack Γ) (ustack Γ) mρ, Σ1) 
    (MkStack (lstack Γ) (ustack Γ) mρ, Σ2)
).
eapply Conf_Unlabel_LValue.
  apply u.
  trivial.
  apply LE_M1M1_cl.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
eapply LowEq_MClosure_Elim.
  apply LE_M1M1.
  apply LE_M1M2_cl.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply IHEV.
  trivial.
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
  apply LE_M1M2.
  apply LE_M2M3.

(* UURefWrite *)
unfold Conf_URefWriteEval_Prop.
intros _t LE_M1M1.
subst ctx.

(* From only touching unlabeled heap *)
assert (LE_M1M2: LowEq_Memory (Γ, Σ1) (Γ, Σ2)). subst.
destruct Σ1.
eapply LowEq_Memory_UHeap_ModR.
  apply LE_M1M1.

apply LE_M1M2.

(* LURefWrite *)
rename H into IHEV. 

unfold Conf_LRefWriteEval_Prop.
intros _t LE_M1M1.
subst ctx.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
eapply IHEV.
  auto.
  assumption.  
  assumption.
Qed.

Lemma Conf_LRefWriteEval : 
  forall LW ctx Γ Σ1 lv Σ2 (EV : LRefWriteEval LW ctx Γ Σ1 lv Σ2), Conf_LRefWriteEval_Prop EV.
Proof.
apply Conf_RefWriteEval.
Qed.

Lemma Conf_URefWriteEval : 
  forall UW ctx Γ Σ1 lv Σ2 (EV : URefWriteEval UW ctx Γ Σ1 lv Σ2), Conf_URefWriteEval_Prop EV.
Proof.
apply Conf_RefWriteEval.
Qed.

(* ******************************************************************************** *)

Lemma LowEq_Fresh_LHeap_UpdateR :
  forall Γ lσ uσ mσ p x v,
    Fresh lσ p ->
    LowEq_Memory (Γ, MkHeap lσ uσ mσ) (Γ, MkHeap lσ uσ mσ) ->
  LowEq_Memory (Γ, MkHeap lσ uσ mσ) (Γ,  MkHeap (Update lσ p (LIFrame (Singleton x (Some v)))) uσ mσ).
Proof.
intros Γ lσ uσ mσ p x v f LE_M1M1.
assert (LE_M1M2 : LowEq_Memory (Γ, MkHeap lσ uσ mσ) (Γ,  MkHeap (Update lσ p (LIFrame (Singleton x (Some v)))) uσ mσ)). 
inversion LE_M1M1 as [_Γ _Γ' _Σ _Σ' LE_LSLS LE_LHLH LE_MSMS LE_MHMH]. subst.
constructor; try assumption.
intros p' p'' bij. simpl.
rewrite (Fresh_Update _ _ f).
apply LE_LHLH.
  apply bij.  

apply LE_M1M2.
Qed.

Lemma LowEq_Fresh_MHeap_UpdateR :
  forall Γ lσ uσ mσ p,
    Fresh mσ p ->
    LowEq_Memory (Γ, MkHeap lσ uσ mσ) (Γ, MkHeap lσ uσ mσ) ->
  LowEq_Memory (Γ, MkHeap lσ uσ mσ) (Γ,  MkHeap  lσ uσ (Update mσ p (MIFrame EmptyMFrame))).
Proof.
intros Γ lσ uσ mσ p f LE_M1M1.
assert (LE_M1M2 : LowEq_Memory (Γ, MkHeap lσ uσ mσ) (Γ,  MkHeap  lσ uσ (Update mσ p (MIFrame EmptyMFrame)))). 
inversion LE_M1M1 as [_Γ _Γ' _Σ _Σ' LE_LSLS LE_LHLH LE_MSMS LE_MHMH]. subst.
constructor; try assumption.
intros p' p'' bij. simpl.
rewrite (Fresh_Update _ _ f).
apply LE_MHMH.
  apply bij.  

apply LE_M1M2.
Qed.


(* ******************************************************************************** *)

Definition Conf_LHeadEval_Prop LHd ctx Γ Σ1 Σ2 lv (EV : LHeadEval LHd ctx Γ Σ1 Σ2 lv) :=
    ctx = H ->
    LowEq_Memory (Γ, Σ1) (Γ, Σ1) ->
  LowEq_Memory (Γ, Σ1) (Γ, Σ2).  


Definition Conf_UHeadEval_Prop UHd ctx Γ Σ1 Σ2 uv (EV : UHeadEval UHd ctx Γ Σ1 Σ2 uv) :=
    ctx = H ->
    LowEq_Memory (Γ, Σ1) (Γ, Σ1) ->
  LowEq_Memory (Γ, Σ1) (Γ, Σ2).  

Lemma Conf_HeadEval : 
  (forall LHd ctx Γ Σ1 Σ2 lv (EV : LHeadEval LHd ctx Γ Σ1 Σ2 lv), Conf_LHeadEval_Prop EV) /\
  (forall UHd ctx Γ Σ1 Σ2 uv (EV : UHeadEval UHd ctx Γ Σ1 Σ2 uv), Conf_UHeadEval_Prop EV).
Proof.

apply HeadEval_mutind with 
  (P  := Conf_LHeadEval_Prop)
  (P0 := Conf_UHeadEval_Prop); intros.

(* LLHead *)
unfold Conf_LHeadEval_Prop.
intros _EQ LE_M1M1. 
assumption.

(* LUHead *)
rename H into IH.
unfold Conf_LHeadEval_Prop.
intros _EQ LE_M1M1.
apply IH.
  trivial.
  apply LE_M1M1.

(* ULHead *)
rename H into IH.
unfold Conf_UHeadEval_Prop.
intros _EQ LE_M1M1. subst.
rewrite Lub_Symm in u. rewrite Lub_Secret in u.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IH.
  trivial.
  apply LE_M1M1.

assert(LE_M2M2:  LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M1M2.

assert (LE_M2M2_cl : LowEq_Memory (MkStack (lstack Γ) (ustack Γ) mρ, Σ2) (MkStack (lstack Γ) (ustack Γ) mρ, Σ2)).
apply LowEq_MClosure_Intro.
  apply LowEq_MStack_Refl.
  apply LE_M2M2.  

assert (LE_M2M3_cl : LowEq_Memory (MkStack (lstack Γ) (ustack Γ) mρ, Σ2) (MkStack (lstack Γ) (ustack Γ) mρ, Σ3)). 
eapply Conf_Unlabel_LValue.
  apply u.
  trivial.
  apply LE_M2M2_cl.

assert (LE_M2M3 :  LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply LowEq_MClosure_Elim.
  apply LE_M2M2.  
  apply LE_M2M3_cl.

eapply LowEq_Memory_Trans.
  apply LE_M1M2.
  apply LE_M2M3.

(* UUHead *)
unfold Conf_UHeadEval_Prop.
intros; assumption.
Qed.

Lemma Conf_LHeadEval : 
  forall LHd ctx Γ Σ1 Σ2 lv (EV : LHeadEval LHd ctx Γ Σ1 Σ2 lv), Conf_LHeadEval_Prop EV.
Proof.
apply Conf_HeadEval.
Qed.

Lemma Conf_UHeadEval : 
  forall UHd ctx Γ Σ1 Σ2 uv (EV : UHeadEval UHd ctx Γ Σ1 Σ2 uv), Conf_UHeadEval_Prop EV.
Proof.
apply Conf_HeadEval.
Qed.

(* ******************************************************************************** *)

Definition Conf_LTailEval_Prop LTl ctx Γ Σ1 Σ2 lv (EV : LTailEval LTl ctx Γ Σ1 Σ2 lv) :=
    ctx = H ->
    LowEq_Memory (Γ, Σ1) (Γ, Σ1) ->
  LowEq_Memory (Γ, Σ1) (Γ, Σ2).


Definition Conf_UTailEval_Prop UTl ctx Γ Σ1 Σ2 uv (EV : UTailEval UTl ctx Γ Σ1 Σ2 uv) :=
    ctx = H ->
    LowEq_Memory (Γ, Σ1) (Γ, Σ1) ->
  LowEq_Memory (Γ, Σ1) (Γ, Σ2).

Lemma Conf_TailEval : 
  (forall LTl ctx Γ Σ1 Σ2 lv (EV : LTailEval LTl ctx Γ Σ1 Σ2 lv), Conf_LTailEval_Prop EV) /\
  (forall UTl ctx Γ Σ1 Σ2 uv (EV : UTailEval UTl ctx Γ Σ1 Σ2 uv), Conf_UTailEval_Prop EV).
Proof.

apply TailEval_mutind with 
  (P  := Conf_LTailEval_Prop)
  (P0 := Conf_UTailEval_Prop); intros.

(* LLTail *)
unfold Conf_LTailEval_Prop.
intros _EQ LE_M1M1. 
assumption.

(* LUTail *)
rename H into IH.
unfold Conf_LTailEval_Prop.
intros _EQ LE_M1M1.
apply IH.
  trivial.
  apply LE_M1M1.

(* ULTail *)
rename H into IH.
unfold Conf_UTailEval_Prop.
intros _EQ LE_M1M1. subst.
rewrite Lub_Symm in u. rewrite Lub_Secret in u.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IH.
  trivial.
  apply LE_M1M1.

assert(LE_M2M2:  LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M1M2.

assert (LE_M2M2_cl : LowEq_Memory (MkStack (lstack Γ) (ustack Γ) mρ, Σ2) (MkStack (lstack Γ) (ustack Γ) mρ, Σ2)).
apply LowEq_MClosure_Intro.
  apply LowEq_MStack_Refl.
  apply LE_M2M2.  

assert (LE_M2M3_cl : LowEq_Memory (MkStack (lstack Γ) (ustack Γ) mρ, Σ2) (MkStack (lstack Γ) (ustack Γ) mρ, Σ3)). 
eapply Conf_Unlabel_LValue.
  apply u.
  trivial.
  apply LE_M2M2_cl.

assert (LE_M2M3 :  LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply LowEq_MClosure_Elim.
  apply LE_M2M2.  
  apply LE_M2M3_cl.

eapply LowEq_Memory_Trans.
  apply LE_M1M2.
  apply LE_M2M3.

(* UUTail *)
unfold Conf_UTailEval_Prop.
intros; assumption.
Qed.

Lemma Conf_LTailEval : 
  forall LTl ctx Γ Σ1 Σ2 lv (EV : LTailEval LTl ctx Γ Σ1 Σ2 lv), Conf_LTailEval_Prop EV.
Proof.
apply Conf_TailEval.
Qed.

Lemma Conf_UTailEval : 
  forall UTl ctx Γ Σ1 Σ2 uv (EV : UTailEval UTl ctx Γ Σ1 Σ2 uv), Conf_UTailEval_Prop EV.
Proof.
apply Conf_TailEval.
Qed.

(* ******************************************************************************** *)

Definition Conf_LFunEval_Prop F ctx Γ Σ1 lv1 Σ2 lv2 (EV : LFunEval F ctx Γ Σ1 lv1 Σ2 lv2) :=
  ctx = H -> 
  LowEq_Memory (Γ, Σ1) (Γ, Σ1) -> 
  LowEq_Memory (Γ, Σ1) (Γ, Σ2).

Definition Conf_UFunEval_Prop F ctx Γ Σ1 uv1 Σ2 uv2 (EV : UFunEval F ctx Γ Σ1 uv1 Σ2 uv2) :=
  ctx = H -> 
  LowEq_Memory (Γ, Σ1) (Γ, Σ1) -> 
  LowEq_Memory (Γ, Σ1) (Γ, Σ2).

Definition Conf_LEval_Prop ctx Γ Σ1 e Σ2 lv (EV : LEval ctx Γ Σ1 e Σ2 lv) := 
  ctx = H -> 
  LowEq_Memory (Γ, Σ1) (Γ, Σ1) -> 
  LowEq_Memory (Γ, Σ1) (Γ, Σ2).

Definition Conf_UEval_Prop ctx Γ Σ1 e Σ2 uv (EV : UEval ctx Γ Σ1 e Σ2 uv) := 
  ctx = H -> 
  LowEq_Memory (Γ, Σ1) (Γ, Σ1) -> 
  LowEq_Memory (Γ, Σ1) (Γ, Σ2).

Lemma Confinement : 
  (forall F ctx Γ Σ1 lv1 Σ2 lv2 (EV : LFunEval F ctx Γ Σ1 lv1 Σ2 lv2) , Conf_LFunEval_Prop EV) /\
  (forall F ctx Γ Σ1 uv1 Σ2 uv2 (EV : UFunEval F ctx Γ Σ1 uv1 Σ2 uv2), Conf_UFunEval_Prop EV) /\
  (forall ctx Γ Σ1 e Σ2 lv (EV : LEval ctx Γ Σ1 e Σ2 lv), Conf_LEval_Prop EV) /\
  (forall ctx Γ Σ1 e Σ2 uv (EV : UEval ctx Γ Σ1 e Σ2 uv), Conf_UEval_Prop EV). 
Proof. 
apply Eval_mutind with 
  (P  := Conf_LFunEval_Prop)
  (P0 := Conf_UFunEval_Prop) 
  (P1 := Conf_LEval_Prop)
  (P2 := Conf_UEval_Prop); intros.

(* ******************************************************************************** *)
(* LLFun *)
rename H into IHEV. 

unfold Conf_LFunEval_Prop.
intros _t LE_M1M1.
subst ctx.
unfold Conf_LEval_Prop in IHEV.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)). subst. 
destruct Σ1.
apply LowEq_Fresh_LHeap_UpdateR.
  apply f.
  apply LE_M1M1.

assert (LE_M1M1_cl : 
  LowEq_Memory 
    (MkStack ρl (ustack Γ) (mstack Γ), Σ1) 
    (MkStack ρl (ustack Γ) (mstack Γ), Σ1)
).
apply LowEq_LClosure_Intro.
  apply LowEq_LStack_Refl.
  apply LE_M1M1.

assert (bij : extension bl p p).
eapply Fresh_LBijection.
  apply f.  
  apply f.

assert (LE_M2M2_cl_ext : 
  LowEq_Memory 
    (MkStack (p :: ρl) (ustack Γ) (mstack Γ), Σ2)
    (MkStack (p :: ρl) (ustack Γ) (mstack Γ), Σ2) 
). 

destruct Σ1. destruct Γ. subst. simpl.
apply Bijection_Memory_LUpdate.
  apply LE_M1M1_cl.
  constructor.
  apply LowEq_Singleton_LFrame.
  apply LowEq_Secret.
  apply bij.

assert (LE_M2M3_cl_ext : 
  LowEq_Memory 
    (MkStack (p :: ρl) (ustack Γ) (mstack Γ), Σ2)
    (MkStack (p :: ρl) (ustack Γ) (mstack Γ), Σ3)
).
apply IHEV.
  trivial.
  apply LE_M2M2_cl_ext.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply LowEq_LClosure_Elim.
  apply LE_M1M1.
  apply LE_M2M3_cl_ext.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
apply LE_M2M3.

(* ******************************************************************************** *)
(* LUFun *) 
rename H into IHEV.

unfold Conf_LFunEval_Prop.
intros _t H.
subst ctx.

assert (LE_M1M2 : LowEq_Memory (Γ1, Σ1) (Γ1, Σ2) ). subst.
destruct Σ1. 
apply LowEq_Fresh_MHeap_UpdateR.
  apply f0.
  assumption.

assert (bij : extension bm p p).
eapply Fresh_MBijection.
  apply f0.  
  apply f0.

assert (LE_M1M2_cl : 
  LowEq_Memory 
    (MkStack (lstack Γ1) (ustack Γ1) mρ, Σ1) 
    (MkStack (lstack Γ1) (ustack Γ1) mρ, Σ2)
).

apply LowEq_MClosure_Intro.
  apply LowEq_MStack_Refl.
  assumption.

assert (LE_M1M1_cl : LowEq_Memory (MkStack (lstack Γ1) (ustack Γ1) mρ, Σ1) (MkStack (lstack Γ1) (ustack Γ1) mρ, Σ1)).
 eapply LowEq_Memory_Refl.
  apply LE_M1M2_cl.

assert (LE_M2M2_cl : LowEq_Memory (MkStack (lstack Γ1) (ustack Γ1)  mρ, Σ2) (MkStack (lstack Γ1) (ustack Γ1) mρ, Σ2)).
 eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M1M2_cl.

assert (LE_M2M2_cl_ext : LowEq_Memory (Γ2, Σ2) (Γ2, Σ2)).
subst. simpl.
apply Bijection_Memory_MUpdate.
  rewrite MkHeap_Equal.
  apply LE_M1M1_cl.
  constructor.
  apply LowEq_EmptyMFrame.
  apply bij.

destruct Γ1 as [lρ1 uρ1 mρ1].
destruct Σ1 as [lσ1 uσ1 mσ1].
simpl in *. 

assert (FrH : Fresh_MStack mρ p).
  eapply Fresh_MPointer_Fresh_MStack.
  apply LE_M1M1_cl.
  assumption.

assert (LE_M2M3_cl_ext : LowEq_Memory (Γ2, Σ2) (Γ2, Σ3)).
eapply Conf_Unlabel_LValue.
  apply u.
  trivial.
  apply LE_M2M2_cl_ext.

assert (LE_M3M3_cl_ext : LowEq_Memory (Γ2, Σ3) (Γ2, Σ3)).
eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M2M3_cl_ext.

assert (LE_M3M3_cl : LowEq_Memory (MkStack lρ1 uρ1 mρ, Σ3) (MkStack lρ1 uρ1 mρ, Σ3)).
inversion LE_M3M3_cl_ext as [_Γ1 _Γ2 _Σ1 _Σ2 LE_LSLS LE_LHLH LE_MSMS LE_MHMH]; subst.
constructor; auto.
subst. inversion LE_MSMS; auto.

assert (LE_M3M4_cl : LowEq_Memory (MkStack lρ1 uρ1 mρ, Σ3) (MkStack lρ1 uρ1 mρ, Σ4)).
eapply Conf_Interp_EffCs.
  eauto.
  apply LE_M3M3_cl.
  subst. apply i.

(* *)

remember (MkStack lρ1 uρ1 mρ1) as Γ1.
remember (MkHeap lσ1 uσ1 mσ1) as Σ1.

assert (LE_M2M3 :  LowEq_Memory (Γ1, Σ2) (Γ1, Σ3)).
eapply LowEq_MClosure_Elim.
  apply H.
  subst. apply LE_M2M3_cl_ext.

assert (LE_M3M4 :  LowEq_Memory (Γ1, Σ3) (Γ1, Σ4)).
eapply LowEq_MClosure_Elim.
  apply LE_M1M2.
  subst. apply LE_M3M4_cl.

assert (LE_M4M4 : LowEq_Memory (Γ1, Σ4) (Γ1, Σ4)).
  eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M3M4.
  
assert (LE_M4M5 :  LowEq_Memory (Γ1, Σ4) (Γ1, Σ5)).
eapply IHEV.
  trivial.
  apply LE_M4M4.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
eapply LowEq_Memory_Trans.
apply LE_M2M3.
eapply LowEq_Memory_Trans.
apply LE_M3M4.
apply LE_M4M5.

(* ******************************************************************************** *)
(* UUFun *)
rename H into IHEV.

unfold Conf_UFunEval_Prop.
intros _t LE_M1M1.
subst ctx. 

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)). subst.
destruct Σ1. simpl.
eapply LowEq_Memory_UHeap_Mod.
apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2) ).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M1M2.

assert (LE_M2M2_cl_ext : 
  LowEq_Memory 
    (MkStack (lstack Γ) (p :: ρu) (mstack Γ), Σ2) 
    (MkStack (lstack Γ) (p :: ρu) (mstack Γ), Σ2)
).
destruct Γ. simpl.
eapply LowEq_Memory_UStack_Mod.
apply LE_M2M2.

assert (LE_M2M3_cl_ext :
  LowEq_Memory  
    (MkStack (lstack Γ) (p :: ρu) (mstack Γ), Σ2) 
    (MkStack (lstack Γ) (p :: ρu) (mstack Γ), Σ3)
).
apply IHEV.
  trivial.
  apply LE_M2M2_cl_ext.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)). subst.
destruct Γ.
eapply LowEq_Memory_UStack_Mod.
  apply LE_M2M3_cl_ext.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
apply LE_M2M3.

(* ******************************************************************************** *)
(* ULFun *) 
rename H into IHEV.

unfold Conf_UFunEval_Prop.
intros _t LE_M1M1.
subst ctx. 

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV.
  auto.
  apply LE_M1M1. 

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply Conf_Unlabel_LValue.
  apply u.  
  trivial.
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
apply LE_M2M3.

(* ******************************************************************************** *)
(* Num *) 
unfold Conf_LEval_Prop.
intros _t LE_M1M1.
apply LE_M1M1.

(* ******************************************************************************** *)
(* Var *) 
unfold Conf_LEval_Prop.
intros _t LE_M1M1.
apply LE_M1M1.

(* ******************************************************************************** *)
(* If1 *)
rename H into IHEV1.
rename H0 into IHEV2.
 
unfold Conf_LEval_Prop.
intros _t LE_M1M1.
subst ctx.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV1.
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
apply IHEV2.
  trivial.
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
apply LE_M2M3.
 
(* ******************************************************************************** *)
(* If2 *) 
rename H into IHEV1.
rename H0 into IHEV2.

unfold Conf_LEval_Prop.
intros _t LE_M1M1.
subst ctx.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV1.
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
apply IHEV2.
  trivial.
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
apply LE_M2M3.

(* ******************************************************************************** *)
(* Seq *) 
rename H into IHEV1.
rename H0 into IHEV2.

unfold Conf_LEval_Prop.
intros _t LE_M1M1.
subst ctx.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV1.
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
apply IHEV2.
  trivial.
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
apply LE_M2M3.

(* ******************************************************************************** *)
(* Ref *) 
rename H into IHEV.

unfold Conf_LEval_Prop.
intros _t LE_M1M1.
subst ctx.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV.
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply Conf_MkLRef.
  apply m.
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
apply LE_M2M3.

(* ******************************************************************************** *)
(* Deref *)
rename H into IHEV.
 
unfold Conf_LEval_Prop.
intros _t LE_M1M1.
subst ctx.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV.
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply Conf_LRefReadEval.
  apply l1.
  auto.  
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
apply LE_M2M3.

(* ******************************************************************************** *)
(* Asn *) 
rename H into IHEV1.
rename H0 into IHEV2. 

unfold Conf_LEval_Prop.
intros _t LE_M1M1.
subst ctx.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV1.
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
apply IHEV2.
  trivial.
  apply LE_M2M2.

assert (LE_M3M3 : LowEq_Memory (Γ, Σ3) (Γ, Σ3)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M2M3.

assert (LE_M3M4 : LowEq_Memory (Γ, Σ3) (Γ, Σ4)).
eapply Conf_LRefWriteEval.
  apply l2.
  auto.
  apply LE_M3M3. 

eapply LowEq_Memory_Trans.
apply LE_M1M2.
eapply LowEq_Memory_Trans.
apply LE_M2M3.
apply LE_M3M4.

(* ******************************************************************************** *)
(* Fun *) 
unfold Conf_LEval_Prop.
intros _t LE_M1M1.
apply LE_M1M1.

(* ******************************************************************************** *)
(* App *)
rename H into IHEV1.
rename H0 into IHEV2.
rename H1 into IHEV3.

unfold Conf_LEval_Prop.
intros _t LE_M1M1.
subst ctx.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV1.
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
apply IHEV2.
  trivial.
  apply LE_M2M2.

assert (LE_M3M3 : LowEq_Memory (Γ, Σ3) (Γ, Σ3)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M2M3.

assert (LE_M3M4 : LowEq_Memory (Γ, Σ3) (Γ, Σ4)).
apply IHEV3.
  trivial.
  apply LE_M3M3.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
eapply LowEq_Memory_Trans.
apply LE_M2M3.
apply LE_M3M4.

(* ******************************************************************************** *)
(* Lib *) 
unfold Conf_LEval_Prop.
intros _t LE_M1M1.
apply LE_M1M1.

(* ******************************************************************************** *)
(* Nil *) 
unfold Conf_LEval_Prop.
intros _t LE_M1M1.
apply LE_M1M1.

(* ******************************************************************************** *)
(* Cons *) 
unfold Conf_LEval_Prop.
intros _t LE_M1M1.
subst ctx.

rename H into IHEV1.
rename H0 into IHEV2.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV1.
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
apply IHEV2.
  trivial.
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
apply LE_M2M3.

(* ******************************************************************************** *)
(* Head *) 
rename H into IH.
unfold Conf_LEval_Prop.
intros _t LE_M1M1. subst.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IH. 
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply Conf_LHeadEval.
  rewrite Lub_Symm in l1. rewrite Lub_Secret in l1. apply l1.
  trivial.
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
  apply LE_M1M2.
  apply LE_M2M3.

(* ******************************************************************************** *)
(* Tail *) 
rename H into IH.
unfold Conf_LEval_Prop.
intros _t LE_M1M1. subst.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IH. 
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply Conf_LTailEval.
  rewrite Lub_Symm in l1. rewrite Lub_Secret in l1. apply l1.
  trivial.
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
  apply LE_M1M2.
  apply LE_M2M3.

(* ******************************************************************************** *)
(* ******************************************************************************** *)
(* ******************************************************************************** *)
(* UNum *) 
unfold Conf_UEval_Prop.
intros _t LE_M1M1.
apply LE_M1M1.

(* ******************************************************************************** *)
(* UVar *) 
unfold Conf_UEval_Prop.
intros _t LE_M1M1.
apply LE_M1M1.

(* ******************************************************************************** *)
(* UVIf1 *) 
rename H into IHEV1.
rename H0 into IHEV2.

unfold Conf_UEval_Prop.
intros _t LE_M1M1.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV1.
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
apply IHEV2.
  trivial.
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
apply LE_M2M3.

(* ******************************************************************************** *)
(* UVIf2 *) 
rename H into IHEV1.
rename H0 into IHEV2.

unfold Conf_UEval_Prop.
intros _t LE_M1M1.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV1.
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
apply IHEV2.
  trivial.
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
apply LE_M2M3.

(* ******************************************************************************** *)
(* UVSeq *) 
rename H into IHEV1.
rename H0 into IHEV2.

unfold Conf_UEval_Prop.
intros _t LE_M1M1.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV1.
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
apply IHEV2.
  trivial.
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
apply LE_M2M3.

(* ******************************************************************************** *)
(* UVRef *)
rename H into IHEV.

unfold Conf_UEval_Prop.
intros _t LE_M1M1.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV.
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply Conf_MkURef.
  apply m.
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
apply LE_M2M3.

(* ******************************************************************************** *)
(* UVDeref *)
rename H into IHEV.

unfold Conf_UEval_Prop.
intros _t LE_M1M1.
subst ctx.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV.
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply Conf_URefReadEval.
  apply u0.
  auto.  
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
apply LE_M2M3.

(* ******************************************************************************** *)
(* UAsn *) 
rename H into IHEV1.
rename H0 into IHEV2.

unfold Conf_UEval_Prop.
intros _t LE_M1M1.
subst ctx.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV1.
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
apply IHEV2.
  trivial.
  apply LE_M2M2.

assert (LE_M3M3 : LowEq_Memory (Γ, Σ3) (Γ, Σ3)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M2M3.

assert (LE_M3M4 : LowEq_Memory (Γ, Σ3) (Γ, Σ4)).
eapply Conf_URefWriteEval.
  apply u1.
  auto.
  apply LE_M3M3. 

eapply LowEq_Memory_Trans.
apply LE_M1M2.
eapply LowEq_Memory_Trans.
apply LE_M2M3.
apply LE_M3M4.

(* ******************************************************************************** *)
(* UFun *) 
unfold Conf_UEval_Prop.
intros _t LE_M1M1.
apply LE_M1M1.

(* ******************************************************************************** *)
(* UApp *) 
rename H into IHEV1.
rename H0 into IHEV2.
rename H1 into IHEV3.

unfold Conf_UEval_Prop.
intros _t LE_M1M1.
subst ctx.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV1.
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
apply IHEV2.
  trivial.
  apply LE_M2M2.

assert (LE_M3M3 : LowEq_Memory (Γ, Σ3) (Γ, Σ3)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M2M3.

assert (LE_M3M4 : LowEq_Memory (Γ, Σ3) (Γ, Σ4)).
apply IHEV3.
  trivial.
  apply LE_M3M3.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
eapply LowEq_Memory_Trans.
apply LE_M2M3.
apply LE_M3M4.

(* ******************************************************************************** *)
(* UNil *) 
unfold Conf_UEval_Prop.
intros _t LE_M1M1.
apply LE_M1M1.

(* ******************************************************************************** *)
(* UCons *) 
rename H into IHEV1.
rename H0 into IHEV2.

unfold Conf_UEval_Prop.
intros _t LE_M1M1.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IHEV1.
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
apply LowEq_Memory_Symm.
apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
apply IHEV2.
  trivial.
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
apply LE_M1M2.
apply LE_M2M3.

(* ******************************************************************************** *)
(* UHead *) 
rename H into IH.
unfold Conf_UEval_Prop.
intros _t LE_M1M1. subst.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IH. 
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply Conf_UHeadEval.
  apply u0.
  trivial.
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
  apply LE_M1M2.
  apply LE_M2M3.

(* ******************************************************************************** *)
(* UTail *) 
rename H into IH.
unfold Conf_UEval_Prop.
intros _t LE_M1M1. subst.

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
apply IH. 
  trivial.
  apply LE_M1M1.

assert (LE_M2M2 : LowEq_Memory (Γ, Σ2) (Γ, Σ2)).
eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M1M2.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply Conf_UTailEval.
  apply u0.
  trivial.
  apply LE_M2M2.

eapply LowEq_Memory_Trans.
  apply LE_M1M2.
  apply LE_M2M3.
Qed.

Lemma Conf_LFunEval : forall F ctx Γ Σ1 lv1 Σ2 lv2 (EV : LFunEval F ctx Γ Σ1 lv1 Σ2 lv2), Conf_LFunEval_Prop EV.
Proof.
apply Confinement.
Qed.

Lemma Conf_UFunEval : forall F ctx Γ Σ1 uv1 Σ2 uv2 (EV : UFunEval F ctx Γ Σ1 uv1 Σ2 uv2), Conf_UFunEval_Prop EV.
Proof.
apply Confinement.
Qed.

Lemma Conf_LEval : forall ctx Γ Σ1 e Σ2 lv (EV : LEval ctx Γ Σ1 e Σ2 lv), Conf_LEval_Prop EV.
Proof.
apply Confinement.
Qed.

Lemma Conf_UEval : forall ctx Γ Σ1 e Σ2 uv (EV : UEval ctx Γ Σ1 e Σ2 uv), Conf_UEval_Prop EV.
Proof.
apply Confinement.
Qed.