Set Implicit Arguments.

Require Import Basics.
Require Import Syntax.
Require Import Semantics.

(* Shortcut the bijection reasoning  *)
Axiom bl : Bijection LPointer.
Axiom bm : Bijection MPointer.

Axiom Fresh_LBijection : 
  forall (σl σl' : LHeap) (p p' : LPointer),
    Fresh σl p -> 
    Fresh σl' p' -> 
    (extension bl) p p'.

Axiom Fresh_MBijection : 
  forall (σm σm' : MHeap) (p p' : MPointer),
    Fresh σm p -> 
    Fresh σm' p' -> 
    (extension bm) p p'.


(* Frames should be in bijection *)
Inductive LowEq_LStack : list LPointer -> list LPointer -> Prop :=
  
  | bLNil : LowEq_LStack nil nil

  | bLCons : 
      forall p1 p2 ps1 ps2, 
        (extension bl) p1 p2 ->
        LowEq_LStack ps1 ps2 ->  
        LowEq_LStack (cons p1 ps1) (cons p2 ps2).
              
Inductive LowEq_MStack : list MPointer -> list MPointer -> Prop :=
  
  | bMNil : LowEq_MStack nil nil

  | bMCons : 
      forall p1 p2 ps1 ps2, 
        (extension bm) p1 p2 ->
        LowEq_MStack ps1 ps2 ->  
        LowEq_MStack (cons p1 ps1) (cons p2 ps2).

(* For the confinement reasoning - all produced MStacks and LStacks have this property *)
Axiom LowEq_MStack_Refl : forall mρ, LowEq_MStack mρ mρ.
Axiom LowEq_LStack_Refl : forall lρ, LowEq_LStack lρ lρ.



(* LowEq_LValue *************** *) 
Inductive LowEq_LValue : LValue -> LValue -> Prop :=
  | leLVNum : 
      forall n, 
        LowEq_LValue (LVNum L n) (LVNum L n)

  | leLVFun : 
      forall F1 F2,
        LowEq_LFun F1 F2 -> 
        LowEq_LValue (LVFun L F1) (LVFun L F2)

  | leLVRef : 
      forall R1 R2 W1 W2,
        LowEq_LRefRead R1 R2 -> 
        LowEq_LRefWrite W1 W2 ->
        LowEq_LValue (LVRef L R1 W1) (LVRef L R2 W2)     

  | leLVNil : 
      LowEq_LValue (LVNil L) (LVNil L)
  
  | leLVCons : 
      forall LHd LHd' LTl LTl',
        LowEq_LHead LHd LHd' ->
        LowEq_LTail LTl LTl' ->
      LowEq_LValue (LVCons L LHd LTl) (LVCons L LHd' LTl')
          
  | leLValueH : 
      forall v1 v2,
        LabelOf v1 = H ->
        LabelOf v2 = H ->
        LowEq_LValue v1 v2

(* LowEq_LFun *************** *) 
with LowEq_LFun : LFun -> LFun -> Prop :=
  | leLLFun : 
      forall lρ1 lρ2 x e, 
        LowEq_LStack lρ1 lρ2 -> 
      LowEq_LFun (LLFun lρ1 x e) (LLFun lρ2 x e)

  | leLUFun :
      forall mρ1 mρ2 UF1 UF2 γ ϕ ξ,
        LowEq_UFun L UF1 UF2 ->
        LowEq_MStack mρ1 mρ2 -> 
    LowEq_LFun (LUFun mρ1 UF1 γ ϕ ξ) (LUFun mρ2 UF2 γ ϕ ξ)

(* LowEq_LRefRead *************** *) 
with LowEq_LRefRead : LRefRead -> LRefRead -> Prop :=
  | leLLRefRead : 
      forall p1 p2,
        (extension bl) p1 p2 ->
      LowEq_LRefRead (LLRefRead p1) (LLRefRead p2)

  | leLURefRead :
      forall mρ1 mρ2 UR1 UR2 γ,
        LowEq_URefRead UR1 UR2 ->
        LowEq_MStack mρ1 mρ2 -> 
      LowEq_LRefRead (LURefRead mρ1 UR1 γ) (LURefRead mρ2 UR2 γ) 

(* LowEq_LRefWrite *************** *) 
with LowEq_LRefWrite : LRefWrite -> LRefWrite -> Prop :=
  | leLLRefWrite : 
      forall p1 p2,
        (extension bl) p1 p2 ->
      LowEq_LRefWrite (LLRefWrite p1) (LLRefWrite p2)

  | leLURefWrite :
      forall mρ1 mρ2 UW1 UW2 γ ϕ,
        LowEq_URefWrite UW1 UW2 ->
        LowEq_MStack mρ1 mρ2 -> 
      LowEq_LRefWrite (LURefWrite mρ1 UW1 γ ϕ) (LURefWrite mρ2 UW2 γ ϕ)

with LowEq_LHead : LHead -> LHead -> Prop :=
  | leLLHead :
      forall lv lv',
        LowEq_LValue lv lv' -> 
      LowEq_LHead (LLHead lv) (LLHead lv')

  | leLUHead : 
      forall mρ1 mρ2 UHd1 UHd2 γ, 
        LowEq_UHead UHd1 UHd2 -> 
        LowEq_MStack mρ1 mρ2 -> 
      LowEq_LHead (LUHead mρ1 UHd1 γ) (LUHead mρ2 UHd2 γ)

with LowEq_LTail : LTail -> LTail -> Prop :=
  | leLLTail :
       forall lv lv',
        LowEq_LValue lv lv' -> 
      LowEq_LTail (LLTail lv) (LLTail lv')   

  | leLUTail : 
      forall mρ1 mρ2 UTl1 UTl2 γ, 
        LowEq_UTail UTl1 UTl2 -> 
        LowEq_MStack mρ1 mρ2 -> 
      LowEq_LTail (LUTail mρ1 UTl1 γ) (LUTail mρ2 UTl2 γ)   

(* LowEq_Uvalue *************** *) 
with LowEq_UValue : Label -> UValue -> UValue -> Prop :=
  | leUVNum : 
      forall n,  
      LowEq_UValue L (UVNum n) (UVNum n)

  | leUVFun : 
      forall UF UF',
        LowEq_UFun L UF UF' -> 
      LowEq_UValue L (UVFun UF) (UVFun UF')

  | leUVRef : 
      forall R R' W W',
        LowEq_URefRead R R' -> 
        LowEq_URefWrite W W' -> 
      LowEq_UValue L (UVRef R W) (UVRef R' W')

  | leUVNil : 
      LowEq_UValue L UVNil UVNil
  
  | leUVCons : 
      forall UHd UHd' UTl UTl',
        LowEq_UHead UHd UHd' ->
        LowEq_UTail UTl UTl' ->
      LowEq_UValue L (UVCons UHd UTl) (UVCons UHd' UTl')

  | leUValueH : 
       forall v1 v2, 
       LowEq_UValue H v1 v2

with LowEq_UFun : Label -> UFun -> UFun -> Prop :=
  | leUUFun : 
      forall uρ1 uρ2 x e,  
      LowEq_UFun L (UUFun uρ1 x e) (UUFun uρ2 x e)
 
  | leULFun : 
      forall LF1 LF2 π,
        LowEq_LFun LF1 LF2 ->
      LowEq_UFun L (ULFun L LF1 π) (ULFun L LF2 π)
       

with LowEq_URefRead : URefRead -> URefRead -> Prop :=
  | leUURefRead : 
      forall p,
      LowEq_URefRead (UURefRead p) (UURefRead p)
 
  | leUKRefRead : 
      forall LR1 LR2 π,
        LowEq_LRefRead LR1 LR2 ->
      LowEq_URefRead (ULRefRead L LR1 π) (ULRefRead L LR2 π)

with LowEq_URefWrite : URefWrite -> URefWrite -> Prop :=
  | leUURefWrite :  
      forall p,
      LowEq_URefWrite (UURefWrite p) (UURefWrite p)

  | leULRefWrite : 
      forall LW1 LW2 π,
        LowEq_LRefWrite LW1 LW2 ->
      LowEq_URefWrite (ULRefWrite L LW1 π) (ULRefWrite L LW2 π)

with LowEq_UHead : UHead -> UHead -> Prop :=
  | leULHead : 
      forall mρ1 mρ2 LHd1 LHd2 ϕ, 
        LowEq_LHead LHd1 LHd2 -> 
        LowEq_MStack mρ1 mρ2 -> 
      LowEq_UHead (ULHead L mρ1 LHd1 ϕ) (ULHead L mρ2 LHd2 ϕ)

  | leUUHead :
      forall uv uv',
        LowEq_UValue L uv uv' -> 
      LowEq_UHead (UUHead uv) (UUHead uv')


with LowEq_UTail : UTail -> UTail -> Prop :=
  | leULTail : 
      forall mρ1 mρ2 LTl1 LTl2 ϕ, 
        LowEq_LTail LTl1 LTl2 -> 
        LowEq_MStack mρ1 mρ2 -> 
      LowEq_UTail (ULTail L mρ1 LTl1 ϕ) (ULTail L mρ2 LTl2 ϕ)

  | leUUTail :
       forall uv uv',
        LowEq_UValue L uv uv' -> 
      LowEq_UTail (UUTail uv) (UUTail uv').

(* LowEq_MValue *************** *) 
Inductive LowEq_MValue : MValue -> MValue -> Prop :=
  | leMValue : 
      forall mv, 
        LowEq_MValue mv mv.
 

(* LowEq_LFrame, LowEq_MFrame *************** *) 
Inductive LowEq_LFrame : LFrame -> LFrame -> Prop :=
  | leLFrame : 
      forall (ω1 ω2 : LFrame),
        (forall x,
          (exists v1 v2, ω1 x = Some v1 /\ ω2 x = Some v2 /\ LowEq_LValue v1 v2)
          \/
          (ω1 x = None /\ ω2 x = None)
        ) ->
        LowEq_LFrame ω1 ω2.

Inductive LowEq_MFrame : MFrame -> MFrame -> Prop :=
  | leMFrame : 
      forall ω1 ω2,
        (forall x,
          (exists mv1 mv2, ω1 x = Some mv1 /\ ω2 x = Some mv2 /\ LowEq_MValue mv1 mv2)
          \/
          (ω1 x = None /\ ω2 x = None)
        ) ->
        LowEq_MFrame ω1 ω2.


(* LowEq_LItem, LowEq_MItem *************** *) 
Inductive LowEq_LItem : LItem -> LItem -> Prop :=
  | leLIValue : 
      forall v1 v2, 
        LowEq_LValue v1 v2 -> 
        LowEq_LItem (LIValue v1) (LIValue v2)

  | leLIFrame : 
      forall ω1 ω2, 
        LowEq_LFrame ω1 ω2 -> 
        LowEq_LItem (LIFrame ω1) (LIFrame ω2).

Inductive LowEq_MItem : MItem -> MItem -> Prop :=
  | leMIFrame : 
      forall ω1 ω2,  
        LowEq_MFrame ω1 ω2 -> 
        LowEq_MItem (MIFrame ω1) (MIFrame ω2).

(* LowEq_Memory *************** *) 
Inductive LowEq_Memory : Stack * Heap -> Stack * Heap -> Prop :=
  | leHeap : 
      forall Γ1 Γ2 Σ1 Σ2,
         LowEq_LStack (lstack Γ1) (lstack Γ2) ->
         (forall p1 p2, 
            (extension bl) p1 p2 -> 
            LowEq_LItem  (lheap Σ1 p1) (lheap Σ2 p2)) ->
         LowEq_MStack (mstack Γ1) (mstack Γ2) ->
         (forall p1 p2,
             (extension bm) p1 p2 ->
             LowEq_MItem (mheap Σ1 p1) (mheap Σ2 p2)) ->
      LowEq_Memory (Γ1, Σ1) (Γ2, Σ2).

(* ******************************************************************************** *)



Axiom LowEq_Memory_Symm : 
  forall Γ1 Γ2 Σ1 Σ2,
    LowEq_Memory (Γ1, Σ1) (Γ2, Σ2) ->
  LowEq_Memory (Γ2, Σ2) (Γ1, Σ1).

Axiom LowEq_Memory_Trans : 
  forall Γ1 Γ2 Γ3 Σ1 Σ2 Σ3,
    LowEq_Memory (Γ1, Σ1) (Γ2, Σ2) ->
    LowEq_Memory (Γ2, Σ2) (Γ3, Σ3) ->
  LowEq_Memory (Γ1, Σ1) (Γ3, Σ3).

Lemma LowEq_Memory_Refl : 
  forall Γ1 Γ2 Σ1 Σ2,
    LowEq_Memory (Γ1, Σ1) (Γ2, Σ2) ->
    LowEq_Memory (Γ1, Σ1) (Γ1, Σ1).
Proof.
intros Γ1 Γ2 Σ1 Σ2 LEM1M2.
eapply LowEq_Memory_Trans.
  apply LEM1M2.
  apply LowEq_Memory_Symm.
  assumption.
Qed.

Lemma LowEq_Secret : forall lv lv', LowEq_LValue (Upgrade H lv) (Upgrade H lv').
Proof.
intros.
unfold Upgrade.
destruct lv, lv'; rewrite Lub_Secret; rewrite Lub_Secret; constructor; unfold LabelOf; auto.
Qed.

Lemma LowEq_Memory_UStack_ModL : 
  forall lρ uρ1 uρ2 mρ lρ' uρ' mρ' Σ Σ',
    LowEq_Memory (MkStack lρ uρ1 mρ, Σ) (MkStack lρ' uρ' mρ', Σ') ->
    LowEq_Memory (MkStack lρ uρ2 mρ, Σ) (MkStack lρ' uρ' mρ', Σ').
Proof.
intros lρ uρ1 uρ2 mρ lρ' uρ' mρ' Σ Σ' LE_M1M1'.
inversion LE_M1M1'. subst.
constructor; assumption.
Qed.

Lemma LowEq_Memory_UStack_ModR : 
  forall lρ uρ mρ lρ' uρ1' uρ2' mρ' Σ Σ',
    LowEq_Memory (MkStack lρ uρ mρ, Σ) (MkStack lρ' uρ1' mρ', Σ') ->
    LowEq_Memory (MkStack lρ uρ mρ, Σ) (MkStack lρ' uρ2' mρ', Σ').
Proof.
intros lρ uρ1 uρ2 mρ lρ' uρ' mρ' Σ Σ' LE_M1M1'.
inversion LE_M1M1'. subst.
constructor; assumption.
Qed.

Lemma LowEq_Memory_UStack_Mod :  
  forall lρ uρ1 uρ2 mρ lρ' uρ1' uρ2' mρ' Σ Σ',
    LowEq_Memory (MkStack lρ uρ1 mρ, Σ) (MkStack lρ' uρ1' mρ', Σ') ->
    LowEq_Memory (MkStack lρ uρ2 mρ, Σ) (MkStack lρ' uρ2' mρ', Σ').
Proof.
intros lρ uρ1 uρ2 mρ lρ' uρ1' uρ2' mρ' Σ Σ' LE_MM'.
inversion LE_MM' as [_Γ _Γ' _Σ _Σ' LE_LSLS' LE_LHLH' LE_MSMS' LE_MHMH']. subst.
constructor; assumption.
Qed.

Lemma LowEq_Memory_UHeap_ModL : 
  forall Γ Γ' lσ uσ mσ lσ' uσ' mσ' uσn, 
    LowEq_Memory (Γ, MkHeap lσ uσ mσ) (Γ', MkHeap lσ' uσ' mσ') ->
    LowEq_Memory (Γ, MkHeap lσ uσn mσ) (Γ', MkHeap lσ' uσ' mσ').
Proof.
intros  Γ Γ' lσ uσ mσ lσ' uσ' mσ' uσn LE_M1M1'.
inversion LE_M1M1'. subst.
constructor; assumption.
Qed.

Lemma LowEq_Memory_UHeap_ModR : 
  forall Γ Γ' lσ uσ mσ lσ' uσ' mσ' uσn', 
    LowEq_Memory (Γ, MkHeap lσ uσ mσ) (Γ', MkHeap lσ' uσ' mσ') ->
    LowEq_Memory (Γ, MkHeap lσ uσ mσ) (Γ', MkHeap lσ' uσn' mσ').
Proof.
intros Γ Γ' lσ uσ mσ lσ' uσ' mσ' uσn' LE_M1M1'.
apply LowEq_Memory_Symm.
eapply LowEq_Memory_UHeap_ModL. 
apply LowEq_Memory_Symm.
apply LE_M1M1'.
Qed.

Lemma LowEq_Memory_UHeap_Mod : 
  forall Γ Γ' σl σu1 σu2 σm σl' σu1' σu2' σm',
    LowEq_Memory (Γ, MkHeap σl σu1 σm) (Γ', MkHeap σl' σu1' σm') ->
  LowEq_Memory (Γ, MkHeap σl σu2 σm) (Γ', MkHeap σl' σu2' σm').
Proof.
intros  Γ Γ' σl σu1 σu2 σm σl' σu1' σu2' σm' LE_MM'.
inversion LE_MM' as [_Γ _Γ' Σ Σ' LE_LSLS' LE_LHLH' LE_MSMS' LE_MHMH']. subst.
constructor; assumption.
Qed.



Lemma LowEq_Singleton_LFrame : 
  forall lv lv' x,  
    LowEq_LValue lv lv' ->
    LowEq_LFrame (Singleton x (Some lv)) (Singleton x (Some lv')).
Proof.
intros lv lv' x LE_lvlv'.
constructor.
intro x'.
destruct (Dec_String x x') as [EQ | EQ].

(* x = x' *)
subst. 
rewrite Singleton_Option_Same, Singleton_Option_Same.
left. 
exists lv, lv'.
split.
  trivial.
split.
  trivial.
  apply LE_lvlv'.

(* x <> x' *)
rewrite (Singleton_Option_Different lv), (Singleton_Option_Different lv').
right.
  split; trivial.
  assumption.
  assumption.
Qed.

Lemma LowEq_EmptyMFrame : LowEq_MFrame EmptyMFrame EmptyMFrame.
Proof.
constructor.
intro x.
unfold EmptyMFrame.
right.
split; trivial.
Qed.



(* ******************************************************************************** *)

Lemma LowEq_MClosure_Intro : 
  forall Γ Σ Γ' Σ' mρ mρ',
    LowEq_MStack mρ mρ' ->
    LowEq_Memory (Γ, Σ) (Γ', Σ') ->
  LowEq_Memory 
    (MkStack (lstack Γ) (ustack Γ) mρ, Σ)
    (MkStack (lstack Γ') (ustack Γ') mρ', Σ').
Proof.
intros Γ Σ Γ' Σ' mρ mρ' LE_mm' LE_MM'.
inversion LE_MM' as [_Γ _Γ' _Σ _Σ' LE_LSLS' LE_LHLH' LE_MSMS' LE_MHMH']. subst.
constructor; assumption.
Qed.

Lemma LowEq_MClosure_Elim : 
  forall Γ Γ' Σ1 Σ1' Σ2 Σ2' mρ mρ',
  LowEq_Memory (Γ, Σ1) (Γ', Σ1') ->
  LowEq_Memory (MkStack (lstack Γ) (ustack Γ) mρ, Σ2) (MkStack (lstack Γ') (ustack Γ') mρ', Σ2') ->
  LowEq_Memory (Γ, Σ2) (Γ', Σ2').
Proof.
intros Γ Γ' Σ1 Σ1' Σ2 Σ2' mρ mρ' LE_M1M1' LE_M2M2'.
inversion LE_M1M1'. subst.
inversion LE_M2M2'. subst.
constructor; assumption.
Qed.

Lemma LowEq_LClosure_Intro : 
  forall Γ Σ Γ' Σ' lρ lρ',
    LowEq_LStack lρ lρ' ->
    LowEq_Memory (Γ, Σ) (Γ', Σ') ->
  LowEq_Memory 
    (MkStack lρ (ustack Γ) (mstack Γ), Σ)
    (MkStack lρ' (ustack Γ') (mstack Γ'), Σ').
Proof.
intros Γ Σ Γ' Σ' mρ mρ' LE_mm' LE_MM'.
inversion LE_MM' as [_Γ _Γ' _Σ _Σ' LE_LSLS' LE_LHLH' LE_MSMS' LE_MHMH']. subst.
constructor; assumption.
Qed.

Lemma LowEq_LClosure_Elim : 
  forall Γ Γ' Σ1 Σ1' Σ2 Σ2' lρ lρ',
  LowEq_Memory (Γ, Σ1) (Γ', Σ1') ->
  LowEq_Memory (MkStack lρ (ustack Γ) (mstack Γ), Σ2) (MkStack lρ' (ustack Γ') (mstack Γ'), Σ2') ->
  LowEq_Memory (Γ, Σ2) (Γ', Σ2').
Proof.
intros Γ Γ' Σ1 Σ1' Σ2 Σ2' mρ mρ' LE_M1M1' LE_M2M2'.
inversion LE_M1M1'. subst.
inversion LE_M2M2'. subst.
constructor; assumption.
Qed.

(* ******************************************************************************** *)


Lemma LowEq_UMemory : 
  forall ρl ρu1 ρu2 ρm ρl' ρu1' ρu2' ρm' σl σu1 σu2 σm σl' σu1' σu2' σm',
    LowEq_Memory 
      (MkStack ρl ρu1 ρm, MkHeap σl σu1 σm) 
      (MkStack ρl' ρu1' ρm', MkHeap σl' σu1' σm') ->
    LowEq_Memory 
      (MkStack ρl ρu2 ρm, MkHeap σl σu2 σm) 
      (MkStack ρl' ρu2' ρm', MkHeap σl' σu2' σm').
Proof.    
intros ρl ρu1 ρu2 ρm ρl' ρu1' ρu2' ρm' σl σu1 σu2 σm σl' σu1' σu2' σm' LE.
inversion LE; subst.
constructor; assumption.
Qed.

(* ******************************************************************************** *)

Lemma Bijection_LHeap_Update :
  forall Γ Γ' σl σu σm σl' σu' σm' v v' p p',
    LowEq_Memory (Γ, MkHeap σl σu σm) (Γ', MkHeap σl' σu' σm') ->
    LowEq_LValue v v' ->
    (extension bl) p p' -> 
  LowEq_Memory (Γ, MkHeap (Update σl p (LIValue v)) σu σm) (Γ', MkHeap (Update σl' p' (LIValue v')) σu' σm').
Proof.
intros Γ Γ' σl σu σm σl' σu' σm' v v' p1 p1' LE_M1M1' LE_vv' bij1.
inversion LE_M1M1' as [_Γ _Γ' Σ Σ' LE_LSLS' LE_LHLH' LE_MSMS' LE_MHMH']. subst.

constructor; try assumption.

intros p2 p2' bij2.
simpl.

destruct (Bijection_Prop _ _ _ _ bl bij1 bij2) as [ [NEQ1 NEQ2] | [EQ1 EQ2] ].
(* p1 <> p2 /\ p1' <> p2' *)
rewrite  (Update_Different σl (LIValue v) NEQ1).
rewrite  (Update_Different σl' (LIValue v') NEQ2).
eapply LE_LHLH'; assumption.
(* p1 = p2 /\ p1' = p2' *)
subst.
rewrite (Update_Same σl p2).
rewrite (Update_Same σl' p2').
constructor.
  apply LE_vv'.
Qed.


Lemma LowEq_Fresh_LHeap_Update : 
  forall Γ Γ' σl σu σm σl' σu' σm' v v' p p',
    LowEq_Memory (Γ, MkHeap σl σu σm) (Γ', MkHeap σl' σu' σm') ->
    LowEq_LValue v v' ->
    Fresh σl p -> 
    Fresh σl' p' -> 
  LowEq_Memory (Γ, MkHeap (Update σl p (LIValue v)) σu σm) (Γ', MkHeap (Update σl' p' (LIValue v')) σu' σm').
Proof.
intros Γ Γ' σl σu σm σl' σu' σm' v v' p1 p1' LE_M1M1' LE_vv' f f'.
inversion LE_M1M1' as [_Γ _Γ' Σ Σ' LE_LSLS' LE_LHLH' LE_MSMS' LE_MHMH']. subst.
assert (bij : extension bl p1 p1'). eapply Fresh_LBijection. apply f. apply f'.
apply Bijection_LHeap_Update; assumption.
Qed.



Lemma Bijection_Memory_LUpdate :
  forall ρl ρu ρm ρl' ρu' ρm' σl σu σm σl' σu' σm' ω ω' p p',
    LowEq_Memory (MkStack ρl ρu ρm, MkHeap σl σu σm) (MkStack ρl' ρu' ρm', MkHeap σl' σu' σm') ->
    LowEq_LItem (LIFrame ω) (LIFrame ω') ->
    (extension bl) p p' -> 
  LowEq_Memory 
    (MkStack (p :: ρl) ρu ρm, MkHeap (Update σl p (LIFrame ω)) σu σm) 
    (MkStack (p' :: ρl') ρu' ρm', MkHeap (Update σl' p' (LIFrame ω')) σu' σm').
Proof.
intros ρl ρu ρm ρl' ρu' ρm' σl σu σm σl' σu' σm' ω ω' p p' LE_M1M1' LE_ff' bij1.
inversion LE_M1M1' as [_Γ _Γ' Σ Σ' LE_LSLS' LE_LHLH' LE_MSMS' LE_MHMH']. subst.

constructor; try assumption.

(* stack *)
constructor; try assumption.

(* heap *)
intros p2 p2' bij2.
simpl.

destruct (Bijection_Prop _ _ _ _ bl bij1 bij2) as [ [NEQ1 NEQ2] | [EQ1 EQ2] ].
(* p1 <> p2 /\ p1' <> p2' *)
rewrite  (Update_Different σl (LIFrame ω) NEQ1).
rewrite  (Update_Different σl' (LIFrame ω') NEQ2).
eapply LE_LHLH'; assumption.
(* p1 = p2 /\ p1' = p2' *)
subst.
rewrite (Update_Same σl p2).
rewrite (Update_Same σl' p2').
constructor.
  inversion LE_ff'. subst. assumption.
Qed.

Lemma Bijection_Memory_MUpdate : 
  forall ρl ρu ρm ρl' ρu' ρm' σl σu σm σl' σu' σm' ω ω' p p',
    LowEq_Memory (MkStack ρl ρu ρm, MkHeap σl σu σm) (MkStack ρl' ρu' ρm', MkHeap σl' σu' σm') ->
    LowEq_MItem (MIFrame ω) (MIFrame ω') -> 
    (extension bm) p p' -> 
  LowEq_Memory
    (MkStack ρl ρu (p :: ρm), MkHeap σl σu (Update σm p (MIFrame ω)))
    (MkStack ρl' ρu' (p' :: ρm'), MkHeap σl' σu' (Update σm' p' (MIFrame ω'))).
Proof.
intros ρl ρu ρm ρl' ρu' ρm' σl σu σm σl' σu' σm' ω ω' p p' LE_M1M1' LE_ff' bij1.
inversion LE_M1M1' as [_Γ _Γ' Σ Σ' LE_LSLS' LE_LHLH' LE_MSMS' LE_MHMH']. subst.

constructor; try assumption.

(* stack *)
constructor; try assumption.

(* heap *)
intros p2 p2' bij2.
simpl.

destruct (Bijection_Prop _ _ _ _ bm bij1 bij2) as [ [NEQ1 NEQ2] | [EQ1 EQ2] ].
(* p1 <> p2 /\ p1' <> p2' *)
rewrite  (Update_Different σm (MIFrame ω) NEQ1).
rewrite  (Update_Different σm' (MIFrame ω') NEQ2).
eapply LE_MHMH'; assumption.
(* p1 = p2 /\ p1' = p2' *)
subst.
rewrite (Update_Same σm p2).
rewrite (Update_Same σm' p2').
constructor.
  inversion LE_ff'. subst. assumption.
Qed.

