Set Implicit Arguments.

Require Import Basics.
Require Import Syntax.
Require Import Semantics.
Require Import LowEq.
Require Import Confinement.
Require Import ModelHypotheses.

(* ******************************************************************************** *)

Lemma NI_LookupL' : 
  forall Γ Γ' Σ Σ' x lvo lvo',
    LowEq_Memory (Γ, Σ) (Γ', Σ') ->
    LookupL (lstack Γ) (lheap Σ) x lvo ->
    LookupL (lstack Γ') (lheap Σ') x lvo' ->
    (exists lv lv', lvo = Some lv /\ lvo' = Some lv' /\ LowEq_LValue lv lv') \/
    (lvo = None /\ lvo' = None).
Proof.
intros Γ Γ' Σ Σ' x lv lv' LE_M1M1'.
inversion LE_M1M1' as [_Γ _Γ' _Σ _Σ' LE_LSLS' LE_LHLH' LE_MSMS' LE_MHMH']. subst.

induction LE_LSLS'.

(* nil *)
intros I I'. 
right.
inversion I; inversion I'. auto.

(* cons *)
intros I I'.
inversion I as [ | lσ _x _p _ps ω v HI1 HI2 | lσ _x _p _ps ω v HI1 HI2 HI3 ]; subst; 
inversion I' as [ | lσ' _x _p _ps ω' v' HI1' HI2' | lσ' _x _p _ps ω' v' HI1' HI2' HI3' ]; subst.

(* Found, Found *)
assert (LE_lili' : LowEq_LItem (lheap Σ p1) (lheap Σ' p2)).
apply LE_LHLH'; apply H. rewrite HI1, HI1' in LE_lili'.
inversion LE_lili' as [ | _ω _ω' LE_ωω' ]. subst.
inversion LE_ωω' as [_ω _ω' LE_LFrameProp ]. subst.

destruct (LE_LFrameProp x) as [ [_v [_v' [HI3 [HI3' LE_lvlv']]]] | [HI3 HI3']].
(* Found *) 
assert (_v = v). congruence.
assert (_v' = v'). congruence. subst.
left.
exists v, v'. 
split. auto. split. auto.
apply LE_lvlv'.

(* Not found *)
assert (C : Some v = None). congruence. discriminate C.

(* Found, Not found *)
assert (LE_lili' : LowEq_LItem (lheap Σ p1) (lheap Σ' p2)).
apply LE_LHLH'; apply H. rewrite HI1, HI1' in LE_lili'.
inversion LE_lili' as [ | _ω _ω' LE_ωω' ]. subst.
inversion LE_ωω' as [_ω _ω' LE_LFrameProp ]. subst.

destruct (LE_LFrameProp x) as [ [_v [_v' [HI3 [HI4' LE_lvlv']]]] | [HI3 HI4']].
assert (C: Some _v = None). congruence. discriminate C.
assert (C: Some v = None). congruence. discriminate C.

(* Not found, Not found *)
assert (LE_lili' : LowEq_LItem (lheap Σ p1) (lheap Σ' p2)).
apply LE_LHLH'; apply H. rewrite HI1, HI1' in LE_lili'.
inversion LE_lili' as [ | _ω _ω' LE_ωω' ]. subst.
inversion LE_ωω' as [_ω _ω' LE_LFrameProp ]. subst.

destruct (LE_LFrameProp x) as [ [_lv [_lv' [HI4 [HI3' LE_lvlv']]]] | [HI4 HI3']].
assert (C: Some _lv = None). congruence. discriminate C.
assert (C: Some lv = None). congruence. discriminate C.

(* Not found, Not found *)
eapply IHLE_LSLS'.
  apply HI3.
  apply HI3'.
Qed.

(* ******************************************************************************** *)

Lemma NI_LookupL : 
  forall Γ Γ' Σ Σ' x lv lv',
    LowEq_Memory (Γ, Σ) (Γ', Σ') ->
    LookupL (lstack Γ) (lheap Σ) x (Some lv) ->
    LookupL (lstack Γ') (lheap Σ') x (Some lv') ->
  LowEq_LValue lv lv'.
Proof.
intros Γ Γ' Σ Σ' x lv lv' LE_M1M1' I I'.
destruct (NI_LookupL' LE_M1M1' I I') as [ [_lv [_lv' [EQ1 [EQ2 LE_lvlv']]]] | [EQ1 EQ2]].

assert (_lv = lv). congruence. 
assert (_lv' = lv'). congruence.
subst.
apply LE_lvlv'.

discriminate EQ1.
Qed.


(* ******************************************************************************** *)

Lemma NI_FindM : 
  forall lρ uρ mρ lσ uσ mσ lρ' uρ' mρ' lσ' uσ' mσ' α x x',
    LowEq_Memory (MkStack lρ uρ mρ, MkHeap lσ uσ mσ) (MkStack lρ' uρ' mρ' , MkHeap lσ' uσ' mσ') ->
    FindM mρ mσ α x -> 
  FindM mρ' mσ' α x' ->  
  (exists p p', x = Some p /\ x' = Some p' /\ extension bm p p') \/
  (x = None /\ x' = None).
Proof.
intros lρ uρ mρ lσ uσ mσ lρ' uρ' mρ' lσ' uσ' mσ' α x x' LE_M1M1' F F'.
inversion LE_M1M1' as [ Γ1 Γ2 Σ1 Σ2 LE_LSLS' LE_LHLH' LE_MSMS' LE_MHMH']. subst. simpl in *.
induction LE_MSMS'.

(* nil *)
inversion F. inversion F'. subst. auto.

(* cons *)
inversion F as [ _mσ _x | _mσ _x p _ps ω v M Fr | _mσ _x p _ps ω v M Fr Fi ]; subst;
inversion F' as [ _mσ' _x' | _mσ' _x' p' _ps' ω' v' M' Fr' | _mσ' _x' p' _ps' ω' v' M' Fr' Fi' ]; subst.

(* base, base *)
left. exists p1, p2.
split. congruence.
split. congruence.
assumption.

(* base, rec *)
assert (LE_MIMI' : LowEq_MItem (mσ p1) (mσ' p2)).
apply LE_MHMH'.
 assumption.
inversion LE_MIMI' as [_ω _ω' LE_FF'].
assert (_ω = ω). congruence. subst. clear H0.
assert (_ω' = ω'). congruence. subst. clear H1.
inversion LE_FF' as [_ω _ω' LE_FF'_prop]. subst.
destruct (LE_FF'_prop α) as [[mv1 [mv2 [EQ1 [EQ2 LE_mv1mv2]]]] | [EQ1 EQ2] ]. 
congruence.
congruence.

(* rec, base *)
assert (LE_MIMI' : LowEq_MItem (mσ p1) (mσ' p2)).
apply LE_MHMH'.
 assumption.
inversion LE_MIMI' as [_ω _ω' LE_FF'].
assert (_ω = ω). congruence. subst. clear H0.
assert (_ω' = ω'). congruence. subst. clear H1.
inversion LE_FF' as [_ω _ω' LE_FF'_prop]. subst.
destruct (LE_FF'_prop α) as [[mv1 [mv2 [EQ1 [EQ2 LE_mv1mv2]]]] | [EQ1 EQ2] ]. 
congruence.
congruence.

(* rec, rec *)
apply IHLE_MSMS'. 
constructor; assumption.
assumption.
assumption.
Qed.

(* ******************************************************************************** *)

Lemma NI_LookupM : 
  forall Γ Γ' Σ Σ' π mi mi',
    LowEq_Memory (Γ, Σ) (Γ', Σ') -> 
    LookupM (mstack Γ) (mheap Σ) π mi ->
    LookupM (mstack Γ') (mheap Σ') π mi' ->
    mi = mi'.
Proof.
intros Γ Γ' Σ Σ' π mi mi' LE_M1M1'.
inversion LE_M1M1' as [_Γ _Γ' _Σ _Σ' LE_LSLS' LE_LHLH' LE_MSMS' LE_MHMH']. subst.

induction LE_MSMS'.

(* nil *)
intros I I'. inversion I. inversion I'. subst.
trivial.

(* cons *) 
intros I I'.

inversion I as [ | lσ _x _p _ps ω v HI1 HI2 | lσ _x _p _ps ω v HI1 HI2 HI3 ]; subst; 
inversion I' as [ | lσ' _x _p _ps ω' v' HI1' HI2' | lσ' _x _p _ps ω' v' HI1' HI2' HI3' ]; subst.

(* Found, Found *)
assert (LE_mimi' : LowEq_MItem (mheap Σ p1) (mheap Σ' p2)).
apply LE_MHMH'.
  apply H.
rewrite HI1, HI1' in LE_mimi'.

inversion LE_mimi' as [ _ω _ω' LE_ωω' ]. subst.
inversion LE_ωω' as [_ω _ω' LE_LFrameProp ]. subst.

destruct (LE_LFrameProp π) as [ [_v [_v' [HI3 [HI3' LE_vv']]]] | [HI3 HI3']].

(* Found *) 
assert (_v = v). congruence.
assert (_v' = v'). congruence. subst.
inversion LE_vv'. subst. auto.

(* Not found *)
assert (C : Some v = None). congruence. discriminate C.

(* Found, Not found *)
assert (LE_mimi' : LowEq_MItem (mheap Σ p1) (mheap Σ' p2)).
apply LE_MHMH'.
  apply H.
rewrite HI1, HI1' in LE_mimi'.

inversion LE_mimi' as [ _ω _ω' LE_ωω' ]. subst.
inversion LE_ωω' as [_ω _ω' LE_LFrameProp ]. subst.

destruct (LE_LFrameProp π) as [ [_v [_v' [HI3 [HI4' LE_lvlv']]]] | [HI3 HI4']].
assert (C: Some _v = None). congruence. discriminate C.
assert (C: Some v = None). congruence. discriminate C.

(* Not found, Not found *)
assert (LE_mimi' : LowEq_MItem (mheap Σ p1) (mheap Σ' p2)).
apply LE_MHMH'.
  apply H.
rewrite HI1, HI1' in LE_mimi'.

inversion LE_mimi' as [ _ω _ω' LE_ωω' ]. subst.
inversion LE_ωω' as [_ω _ω' LE_LFrameProp ]. subst.

destruct (LE_LFrameProp π) as [ [_v [_v' [HI4 [HI3' LE_lvlv']]]] | [HI4 HI3']].
assert (C: Some _v = None). congruence. discriminate C.
assert (C: Some v' = None). congruence. discriminate C.

(* Not found, Not found *)
eapply IHLE_MSMS'.
  apply HI3.
  apply HI3'.
Qed.

(* ******************************************************************************** *)

Lemma NI_UpdateM :
  forall lρ uρ mρ lσ uσ mσ1 lρ' uρ' mρ' lσ' uσ' mσ1' mσ2 mσ2' π u, 
    LowEq_Memory (MkStack lρ uρ mρ, MkHeap lσ uσ mσ1) (MkStack lρ' uρ' mρ', MkHeap lσ' uσ' mσ1') ->
    UpdateM L mρ mσ1 π u mσ2 ->
    UpdateM L mρ' mσ1' π u mσ2' ->
 LowEq_Memory (MkStack lρ uρ mρ, MkHeap lσ uσ mσ2) (MkStack lρ' uρ' mρ', MkHeap lσ' uσ' mσ2').
Proof.
intros lρ uρ mρ lσ uσ mσ1 lρ' uρ' mρ' lσ' uσ' mσ1' mσ2 mσ2' π l LE_M1M1' U U'.
inversion LE_M1M1' as [ Γ1 Γ2 Σ1 Σ2 LE_LSLS' LE_LHLH' LE_MSMS' LE_MHMH']. subst. simpl in *.

constructor; auto.
(* MHeap *)
inversion U as [ p1 ps _mσ1 _mσ2 α _l ω1 ω2 FM HH UF UH | ctx _mρ p1 _mσ1 _mσ2 α l1 l2 ω1 ω2 FM HH FH CF UF UH]; subst;
inversion U' as [ p1' ps' _mσ1 _mσ2 α' _l ω1' ω2' FM' HH' UF' UH' | ctx' _mρ p1' _mσ1 _mσ2 α' l1' l2' ω1' ω2' FM' HH' FH' CF' UF' UH']; subst.

(* No, No - same proof as NI_DefineM *)
inversion LE_MSMS' as [ | _p1 _p1' _ps _ps' bij1 LE_MSMS'_sub]. subst.

(* Establish that LowEq_MFrame ω1 ω1' *)
assert (LE_MIMI' : LowEq_MItem (mσ1 p1) (mσ1' p1')).
apply LE_MHMH'.
 apply bij1.

inversion LE_MIMI' as [_ω1 _ω1' LE_FrFr' H0 H1]. subst.
assert (_ω1 = ω1). congruence.
assert (_ω1' = ω1'). congruence.
subst. clear H0 H1.

(* Establish that LowEq_MFrame (Update ω1 π (Some (MVLavel l))) (Update ω1' π (Some (MVLabel l))) *)
assert (LE_FrFr'_Upd : LowEq_MFrame (Update ω1 π (Some (MVLabel l))) (Update ω1' π (Some (MVLabel l)))).
constructor.
intro x.
destruct (Dec_String π x) as [EQ | EQ].

(* π = x *)
subst. 
rewrite Update_Same, Update_Same.
left. exists (MVLabel l), (MVLabel l).
split. trivial. split. trivial.
constructor.
(* π <> c *)
rewrite (Update_Different _ _ EQ), (Update_Different _ _ EQ).
inversion LE_FrFr' as [_ω1 _ω1' LE_FrFr'_Prop]. subst.
apply LE_FrFr'_Prop.

(* Prove that updated heaps are low equivalent *)
intros p2 p2' bij2.
simpl.

destruct (Bijection_Prop _ _ _ _ _ bij1 bij2) as [[EQ1 EQ2] | [EQ1 EQ2]].
(* p1 <> p2 /\ p1' <> p2' *)
rewrite (Update_Different _ _ EQ1), (Update_Different _ _ EQ2).
apply LE_MHMH'.
assumption.

(* p1 = p2 /\ p1' = p2' *)
subst.
rewrite Update_Same, Update_Same.
constructor.
assumption.

(* No, Yes *)
destruct (NI_FindM LE_M1M1' FM FM') as [ [_p [_p' [EQ1 [EQ2 EQ3]]]] | [EQ1 EQ2]]; congruence.
(* Yes, No *)
destruct (NI_FindM LE_M1M1' FM FM') as [ [_p [_p' [EQ1 [EQ2 EQ3]]]] | [EQ1 EQ2]]; congruence.

(* Yes, Yes *)
destruct (NI_FindM LE_M1M1' FM FM') as [ [_p [_p' [EQ1 [EQ2 bij1]]]] | [EQ1 EQ2]]; try congruence.
assert (_p = p1). congruence.
assert (_p' = p1'). congruence.
subst. clear EQ1 EQ2.

(* Establish that LowEq_MFrame ω1 ω1' *)
assert (LE_MIMI' : LowEq_MItem (mσ1 p1) (mσ1' p1')).
apply LE_MHMH'.
 apply bij1.

inversion LE_MIMI' as [_ω1 _ω1' LE_FrFr' H0 H1]. subst.
assert (_ω1 = ω1). congruence.
assert (_ω1' = ω1'). congruence.
subst. clear H0 H1.

(* Establish that l2 = l2' *)
inversion LE_FrFr' as [_ω1 _ω1' LE_FrFr'_Prop]. subst.
destruct (LE_FrFr'_Prop π) as [ [mv [mv' [EQ1 [EQ2 LE_l2l2']]]] | [EQ1 EQ2]]; try congruence.
assert (mv = MVLabel l2). congruence.
assert (mv' = MVLabel l2'). congruence.
subst. clear EQ1 EQ2.
inversion LE_l2l2'. subst l2'. subst.


(* Establish that LowEq_MFrame (Update ω1 π (Some (MVLavel l))) (Update ω1' π (Some (MVLabel l))) *)
assert (LE_FrFr'_Upd : LowEq_MFrame (Update ω1 π (Some (MVLabel (lub l l2)))) (Update ω1' π (Some (MVLabel (lub l l2))))).
constructor.
intro x.
destruct (Dec_String π x) as [EQ | EQ].

(* π = x *)
subst. 
rewrite Update_Same, Update_Same.
left. exists (MVLabel (lub l l2)), (MVLabel (lub l l2)).
split. trivial. split. trivial.
constructor.
(* π <> c *)
rewrite (Update_Different _ _ EQ), (Update_Different _ _ EQ).
apply LE_FrFr'_Prop.

(* Prove that updated heaps are low equivalent *)
intros p2 p2' bij2.
simpl.

destruct (Bijection_Prop _ _ _ _ _ bij1 bij2) as [[EQ1 EQ2] | [EQ1 EQ2]].
(* p1 <> p2 /\ p1' <> p2' *)
rewrite (Update_Different _ _ EQ1), (Update_Different _ _ EQ2).
apply LE_MHMH'.
assumption.

(* p1 = p2 /\ p1' = p2' *)
subst.
rewrite Update_Same, Update_Same.
constructor.
assumption.
Qed.

(* ******************************************************************************** *)

Lemma NI_DefineM :
  forall lρ uρ mρ lσ uσ mσ1 lρ' uρ' mρ' lσ' uσ' mσ1' mσ2 mσ2' π u, 
    LowEq_Memory (MkStack lρ uρ mρ, MkHeap lσ uσ mσ1) (MkStack lρ' uρ' mρ', MkHeap lσ' uσ' mσ1') ->
    DefineM mρ mσ1 π u mσ2 ->
    DefineM mρ' mσ1' π u mσ2' ->
 LowEq_Memory (MkStack lρ uρ mρ, MkHeap lσ uσ mσ2) (MkStack lρ' uρ' mρ', MkHeap lσ' uσ' mσ2').
Proof.
intros lρ uρ mρ lσ uσ mσ1 lρ' uρ' mρ' lσ' uσ' mσ1' mσ2 mσ2' π u LE_M1M1' D D'.
inversion LE_M1M1' as [ Γ1 Γ2 Σ1 Σ2 LE_LSLS' LE_LHLH' LE_MSMS' LE_MHMH']. subst. simpl in *.

constructor; auto.
(* MHeap *)
inversion D as [p1 ps σ1 σ2 x m ω1 ω2 HH NH UF UH]; subst.
inversion D' as [p1' ps' σ1' σ2' x' m' ω1' ω2' HH' NH' UF' UH']; subst.

inversion LE_MSMS' as [ | _p1 _p1' _ps _ps' bij1 LE_MSMS'_sub]. subst.

(* Establish that LowEq_MFrame ω1 ω1' *)
assert (LE_MIMI' : LowEq_MItem (mσ1 p1) (mσ1' p1')).
apply LE_MHMH'.
 apply bij1.

inversion LE_MIMI' as [_ω1 _ω1' LE_FrFr' H0 H1]. subst.
assert (_ω1 = ω1). congruence.
assert (_ω1' = ω1'). congruence.
subst. clear H0 H1.

(* Establish that LowEq_MFrame (Update ω1 π (Some u)) (Update ω1' π (Some u)) *)
assert (LE_FrFr'_Upd : LowEq_MFrame (Update ω1 π (Some u)) (Update ω1' π (Some u))).
constructor.
intro x.
destruct (Dec_String π x) as [EQ | EQ].

(* π = x *)
subst. 
rewrite Update_Same, Update_Same.
left. exists u, u.
split. trivial. split. trivial.
constructor.
(* π <> c *)
rewrite (Update_Different _ _ EQ), (Update_Different _ _ EQ).
inversion LE_FrFr' as [_ω1 _ω1' LE_FrFr'_Prop]. subst.
apply LE_FrFr'_Prop.

(* Prove that updated heaps are low equivalent *)
intros p2 p2' bij2.
simpl.

destruct (Bijection_Prop _ _ _ _ _ bij1 bij2) as [[EQ1 EQ2] | [EQ1 EQ2]].
(* p1 <> p2 /\ p1' <> p2' *)
rewrite (Update_Different _ _ EQ1), (Update_Different _ _ EQ2).
apply LE_MHMH'.
assumption.

(* p1 = p2 /\ p1' = p2' *)
subst.
rewrite Update_Same, Update_Same.
constructor.
assumption.
Qed.

(* ******************************************************************************** *)

Lemma NI_Interp_LTerm : 
  forall Γ Σ Γ' Σ' κ l l',
    LowEq_Memory (Γ, Σ) (Γ', Σ') ->
    Interp_LTerm (mstack Γ) (mheap Σ) κ l ->
    Interp_LTerm (mstack Γ') (mheap Σ') κ l' ->
  l = l'.
Proof.
intros Γ Σ Γ' Σ' κ. 
induction κ. 

(* LVar s *)
intros l l' LE_M1M1' I I'.
inversion I as [ | ρ σ α LH | ρ σ α _l LH | ]; subst s; subst;
inversion I' as [ | ρ' σ' α' LH' | ρ' σ' α' _l' LH' | ]; subst.

(* not found, not found *)
trivial.
(* found, not found *)
discriminate (NI_LookupM LE_M1M1' LH LH').
(* not found, found *)
discriminate (NI_LookupM LE_M1M1' LH LH').
(* found, found *)
assert (EQ : Some (MVLabel l) = Some (MVLabel l')).
apply (NI_LookupM LE_M1M1' LH LH').
congruence.

(* lub *)
intros l l' LE_M1M1' I I'.
inversion I as [ | | | ρ σ _κ1 _κ2 l1 l2 I1 I2]. subst.
inversion I' as [ | | | ρ' σ' _κ1 _κ2 l1' l2' I1' I2']. subst.

assert (l1 = l1').
apply IHκ1; assumption.
assert (l2 = l2').
apply IHκ2; assumption.
congruence.

(* LLabel *)
intros l1 l1' LE_M1M1' I I'.
inversion I. 
inversion I'.
congruence.
Qed.

(* ******************************************************************************** *)

Lemma NI_Interp_EffC : 
  forall Γ Σ1 Γ' Σ1' l Σ2 Σ2' c,
    LowEq_Memory (Γ, Σ1) (Γ', Σ1') ->
    Interp_EffC l Γ Σ1 c Σ2 ->
    Interp_EffC l Γ' Σ1' c Σ2' ->
  LowEq_Memory (Γ, Σ2) (Γ', Σ2').
Proof.
intros Γ Σ1 Γ' Σ1' l Σ2 Σ2' c.
induction c.

(* ECRefRead *)
intros LE_M1M1' I I'.
destruct Γ, Σ1, Γ', Σ1'.

inversion I.
inversion I'.
subst.
eapply NI_DefineM.
  apply LE_M1M1'.
  eauto.
  eauto.

(* ECRefWrite *)
intros LE_M1M1' I I'.
destruct Γ, Σ1, Γ', Σ1'.

inversion I.
inversion I'.
subst.
eapply NI_DefineM.
  apply LE_M1M1'.
  eauto.
  eauto.

(* ECApp *)
intros LE_M1M1' I I'.
destruct Γ, Σ1, Γ', Σ1'.

inversion I.
inversion I'.
subst.
eapply NI_DefineM.
  apply LE_M1M1'.
  eauto.
  eauto.

(* ECAsn *)
intros LE_M1M1' I I'.
rename c into cκ. rename l into κ.
inversion I as [ | | | lρ uρ mρ lσ uσ mσ1 mσ2 ctx α _cκ _κ cl l I1 I2 U]. subst.
inversion I' as [ | | | lρ' uρ' mρ' lσ' uσ' mσ1' mσ2' ctx' _α _cκ _κ cl' l' I1' I2' U']. subst.

assert (cl = cl'). 
eapply NI_Interp_LTerm; eauto.
assert (l = l').
eapply NI_Interp_LTerm; eauto.
subst l' cl'.
simpl in *.

destruct (lub κ cl).
(* lub κ c = L *)
eapply NI_UpdateM; eauto.

(* lub κ c = H *)
assert (LE_M1M2 : LowEq_Memory (MkStack lρ uρ mρ, MkHeap lσ uσ mσ1) (MkStack lρ uρ mρ, MkHeap lσ uσ mσ2)).
eapply Conf_UpdateM; eauto.
  eapply LowEq_Memory_Refl; eauto.

assert (LE_M1'M2' : LowEq_Memory (MkStack lρ' uρ' mρ', MkHeap lσ' uσ' mσ1') (MkStack lρ' uρ' mρ', MkHeap lσ' uσ' mσ2')).
eapply Conf_UpdateM; eauto.
  eapply LowEq_Memory_Refl.
  eapply LowEq_Memory_Symm.
  eauto.

eapply LowEq_Memory_Trans.
  apply LowEq_Memory_Symm.  eauto.
  eapply LowEq_Memory_Trans. apply LE_M1M1'. auto.
Qed.  

(* ******************************************************************************** *)

Lemma NI_Interp_EffCs : 
  forall ξ Γ Σ1 Γ' Σ1' l Σ2 Σ2' ,
    LowEq_Memory (Γ, Σ1) (Γ', Σ1') ->
    Interp_EffCs l Γ Σ1 ξ Σ2 ->
    Interp_EffCs l Γ' Σ1' ξ Σ2' ->
  LowEq_Memory (Γ, Σ2) (Γ', Σ2').
Proof.
induction ξ.

(* nil *)
intros Γ Σ1 Γ' Σ1' l Σ3 Σ3' LE_M1M1' IE IE'.
inversion IE. inversion IE'. subst.
apply LE_M1M1'.

(* cons *)
intros Γ Σ1 Γ' Σ1' l Σ3 Σ3' LE_M1M1' IE IE'.

inversion IE as [ | ctx  _Γ _Σ1 Σ2 _Σ3 c cs IEc IEcs]. subst.
inversion IE' as [ | ctx' _Γ' _Σ1' Σ2' _Σ3' c' cs' IEc' IEcs']. subst.

assert (LE_M2M2' : LowEq_Memory (Γ, Σ2) (Γ', Σ2')).
eapply NI_Interp_EffC.
  apply LE_M1M1'.
  apply IEc.
  apply IEc'.  

assert (LE_M3M3' :  LowEq_Memory (Γ, Σ3) (Γ', Σ3')).
eapply IHξ.
  apply LE_M2M2'.
  apply IEcs.
  apply IEcs'.

apply LE_M3M3'.
Qed.

(* ******************************************************************************** *)

Lemma LowEq_MkLRef : 
  forall Γ Γ' Σ1 Σ1' Σ2 Σ2' v v' LR LR' LW LW', 
    LowEq_Memory (Γ, Σ1) (Γ', Σ1') ->
    LowEq_LValue v v' ->
    MkLRef Σ1 v (Σ2, LR, LW) -> 
    MkLRef Σ1' v' (Σ2', LR', LW') ->
  LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\ LowEq_LValue (LVRef L LR LW) (LVRef L LR' LW').
Proof.
intros Γ Γ' Σ1 Σ1' Σ2 Σ2' v v' LR LR' LW LW' LE_M1M1' LE_vv' MK MK'.

inversion MK as [ σl σu σm lv p f ]. subst.
inversion MK' as [ σl' σu' σm' lv' p' f' ]. subst.


assert (LE_M2M2' : LowEq_Memory (Γ, MkHeap (Update σl p (LIValue v)) σu σm) (Γ', MkHeap (Update σl' p' (LIValue v')) σu' σm')).
apply LowEq_Fresh_LHeap_Update; assumption.

split.
  apply LE_M2M2'.
  constructor.
  constructor. 
  eapply Fresh_LBijection. apply f. apply f'.
  constructor. 
  eapply Fresh_LBijection. apply f. apply f'.
Qed.

(* ******************************************************************************** *)

Lemma NI_LHeap_Lookup_LValue : 
  forall Γ Σ Γ' Σ' p p' lv lv',
    extension bl p p' -> 
    LowEq_Memory (Γ, Σ) (Γ', Σ') ->
    lheap Σ p = LIValue lv ->
    lheap Σ' p' = LIValue lv' ->
 LowEq_LValue lv lv'.
Proof.
intros Γ Σ Γ' Σ' p p' lv lv' bij LE_M1M1' H1 H2.
inversion LE_M1M1' as [_Γ _Γ' _Σ _Σ' LE_LSLS' LE_LHLH' LE_MSMS' LE_MHMH']. subst.

assert (LE_LILI' : LowEq_LItem (lheap Σ p) (lheap Σ' p')).
apply LE_LHLH'.
  apply bij.

inversion LE_LILI' as [_lv _lv' LE_lvlv' | ].
(* LValue *)
assert (_lv = lv). congruence.
assert (_lv' = lv'). congruence.
subst.

apply LE_lvlv'.

(* LFrame *)
assert (C: LIValue lv = LIFrame ω1). congruence. 
discriminate C.
Qed.

(* ******************************************************************************** *)

Lemma NI_MkLRef : 
  forall Γ Γ' Σ1 Σ1' Σ2 Σ2' v v' LR LR' LW LW', 
    LowEq_Memory (Γ, Σ1) (Γ', Σ1') ->
    LowEq_LValue v v' ->
    MkLRef Σ1 v (Σ2, LR, LW) -> 
    MkLRef Σ1' v' (Σ2', LR', LW') ->
  LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\ LowEq_LValue (LVRef L LR LW) (LVRef L LR' LW').
Proof.
intros Γ Γ' Σ1 Σ1' Σ2 Σ2' v v' LR LR' LW LW' LE_M1M1' LE_vv' MK MK'.

inversion MK as [ σl σu σm lv p f ]. subst.
inversion MK' as [ σl' σu' σm' lv' p' f' ]. subst.


assert (LE_M2M2' : LowEq_Memory (Γ, MkHeap (Update σl p (LIValue v)) σu σm) (Γ', MkHeap (Update σl' p' (LIValue v')) σu' σm')).
apply LowEq_Fresh_LHeap_Update; assumption.

split.
  apply LE_M2M2'.
  constructor.
  constructor. 
  eapply Fresh_LBijection. apply f. apply f'.
  constructor. 
  eapply Fresh_LBijection. apply f. apply f'.
Qed.



(* ******************************************************************************** *)
Lemma LowEq_MStack_Top :
  forall Γ Γ' p p' ps ps',
    LowEq_MStack (mstack Γ) (mstack Γ') ->
    cons p ps = mstack Γ->
    cons p' ps' = mstack Γ' ->
    extension bm p p'.
Proof.
intros Γ Γ' p p' ps ps' LE_MSMS' EQ EQ'.
rewrite <- EQ in LE_MSMS'.
rewrite <- EQ' in LE_MSMS'.
inversion LE_MSMS'.
assumption.
Qed.

Lemma LowEq_MFrame_Update : 
  forall ω1 ω1' α x, 
    LowEq_MFrame ω1 ω1' ->
  LowEq_MFrame (Update ω1 α (Some x)) (Update ω1' α (Some x)).
Proof.
intros ω1 ω1' α x LE_FrFr'.
constructor.
intros α'.
destruct (Dec_String α α') as [EQ | EQ].

subst.
left. exists x, x.
rewrite Update_Same, Update_Same.
split. auto. split. auto. 
constructor.

rewrite (Update_Different _ _ EQ), (Update_Different _ _ EQ).
inversion LE_FrFr' as [_ω1 _ω1' LE_FrFr'_Prop]. subst.
apply LE_FrFr'_Prop.
Qed.

Lemma Unlabel_LValue_Heap :
  forall Γ Σ1 Σ2 ϕ lv uv,
    Unlabel_LValue L Γ Σ1 lv ϕ uv Σ2 ->
  exists mσ, UpdateM L (mstack Γ) (mheap Σ1) (LVarOf ϕ) (LabelOf lv) mσ /\ Σ2 = MkHeap (lheap Σ1) (uheap Σ1) mσ.
Proof.
intros Γ Σ1 Σ2 ϕ lv uv U.
inversion U; subst; exists mσ; simpl; split; auto.
Qed.

Lemma NI_Unlabel_LValue : 
  forall Γ Γ' Σ1 Σ1' Σ2 Σ2' ϕ lv uv lv' uv',
    LowEq_Memory (Γ, Σ1) (Γ', Σ1') ->
    LowEq_LValue lv lv' -> 
    Unlabel_LValue L Γ Σ1 lv ϕ uv Σ2 ->
    Unlabel_LValue L Γ' Σ1' lv' ϕ uv' Σ2' ->
  LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\ 
  exists l, LowEq_UValue l uv uv'.
Proof.
intros  Γ Γ' Σ1 Σ1' Σ2 Σ2' ϕ lv uv lv' uv' LE_M1M1' LE_lvlv' UL UL'.
inversion LE_lvlv' as [ n | F F' LE_FF' | R R' W W' LE_RR' LE_WW' | | | v v' L L' ].

(* LVNum *)
subst.
inversion UL as [ ctx _Γ _Σ1 _Σ2 mσ l _n α H1 H2 | | | | ]. subst.
inversion UL' as [ ctx' _Γ' _Σ1' _Σ2' mσ' l' _n' α' H1' H2' | | | | ]. subst.

split.
destruct Γ, Γ', Σ1, Σ1'.
apply (NI_UpdateM LE_M1M1' H1 H1'). 
exists L. constructor.

(* LVFun *)
subst.
inversion UL as [ | ctx _Γ _Σ1 _Σ2 mσ l LF α π H1 H2 | | | ]. subst.
inversion UL' as [ | ctx' _Γ' _Σ1' _Σ2' mσ' l' LF' α' π' H1' H2' | | | ]. subst.

split.
destruct Γ, Γ', Σ1, Σ1'.
apply (NI_UpdateM LE_M1M1' H1 H1'). 
exists L. constructor. constructor. apply LE_FF'.

(* LVRef *)
subst.
inversion UL as [ | | ctx _Γ _Σ1 _Σ2 mσ l LR LW α π H1 H2 | | ]. subst.
inversion UL' as [ | | ctx' _Γ' _Σ1' _Σ2' mσ' l' LR' LW' α' π' H1' H2' | | ]. subst.

split.
destruct Γ, Γ', Σ1, Σ1'.
apply (NI_UpdateM LE_M1M1' H1 H1'). 
exists L.
  constructor.
    constructor.
    apply LE_RR'.
    constructor.
    apply LE_WW'.

(* LVNil *)
subst.
inversion UL as [ | | | ctx _Γ _Σ1 _Σ2 mσ l α eϕ H1 H2 | ]. subst.
inversion UL' as [ | | | ctx' _Γ' _Σ1' _Σ2' mσ' l' α' eϕ' H1' H2' | ]. subst.

split.
destruct Γ, Γ', Σ1, Σ1'.
apply (NI_UpdateM LE_M1M1' H1 H1'). 
exists L.
  constructor.

(* LVCons *) 
subst.
inversion UL as [ | | | | ctx _Γ _Σ1 _Σ2 mσ l Hd Tl eα eϕ H1 H2 ]. subst.
inversion UL' as [ | | | | ctx' _Γ' _Σ1' _Σ2' mσ' l' Hd' Tl' eα' eϕ' H1' H2' ]. subst.

split.
destruct Γ, Γ', Σ1, Σ1'.
apply (NI_UpdateM LE_M1M1' H1 H1'). 
exists L.
  constructor.
  constructor. assumption.
  inversion LE_M1M1'. assumption.

  constructor. assumption.
  inversion LE_M1M1'. assumption.
  
(* LE Secret *)
subst;

split.

assert  (EX_m : exists mσ, UpdateM Syntax.L (mstack Γ) (mheap Σ1) (LVarOf ϕ) (LabelOf lv) mσ /\ 
                           Σ2 = MkHeap (lheap Σ1) (uheap Σ1) mσ).
eapply Unlabel_LValue_Heap.
  apply UL.

assert  (EX_m' : exists mσ', UpdateM Syntax.L (mstack Γ') (mheap Σ1') (LVarOf ϕ) (LabelOf lv') mσ' /\ 
                           Σ2' = MkHeap (lheap Σ1') (uheap Σ1') mσ').
eapply Unlabel_LValue_Heap.
  apply UL'.

destruct EX_m as [mσ [U EQ]].
destruct EX_m' as [mσ' [U' EQ']].
subst. rewrite L in U. rewrite L' in U'.
destruct Γ, Γ', Σ1, Σ1'.
apply (NI_UpdateM LE_M1M1' U U'). 

exists H.
  constructor.
Qed.
   
(* ******************************************************************************** *)

Definition NI_LRefReadEval_Prop LR ctx Γ Σ1 Σ2 lv (EV : LRefReadEval LR ctx Γ Σ1 Σ2 lv) :=
  forall LR' Γ' Σ1' lv' Σ2',
  ctx = L ->
    LowEq_Memory (Γ, Σ1) (Γ', Σ1') ->
    LowEq_LRefRead LR LR' ->
    LRefReadEval LR' ctx Γ' Σ1' Σ2' lv' ->
  LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\
  LowEq_LValue lv lv'.


Definition NI_URefReadEval_Prop UR ctx Γ Σ1 Σ2 uv (EV : URefReadEval UR ctx Γ Σ1 Σ2 uv) :=
 forall UR' Γ' Σ1' uv' Σ2',
  ctx = L ->
    LowEq_Memory (Γ, Σ1) (Γ', Σ1') ->
    LowEq_URefRead UR UR' ->
    URefReadEval UR' ctx Γ' Σ1' Σ2' uv' ->
  LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\
  exists l, LowEq_UValue l uv uv'.

Lemma NI_RefReadEval :
  (forall LR ctx Γ Σ1 Σ2 lv (EV : LRefReadEval LR ctx Γ Σ1 Σ2 lv), NI_LRefReadEval_Prop EV) /\ 
  (forall UR ctx Γ Σ1 Σ2 uv (EV : URefReadEval UR ctx Γ Σ1 Σ2 uv), NI_URefReadEval_Prop EV).
Proof.  
apply RefReadEval_mutind with
  (P  := NI_LRefReadEval_Prop)
  (P0 := NI_URefReadEval_Prop); intros.

(* LLRefRead *)
unfold NI_LRefReadEval_Prop.
intros LR' Γ' Σ1' lv' Σ2' _t LE_M1M1' LE_LRLR' EV'.

inversion LE_LRLR' as [_p p' bij | ]. subst.

inversion EV' as [ctx' _Γ' Σ' _p' _lv' e' | ]. subst Σ1' Σ2'. subst.

assert (LE_lvlv' : LowEq_LValue lv lv').
eapply NI_LHeap_Lookup_LValue.
  apply bij.
  apply LE_M1M1'.
  apply e.
  apply e'.

split.
  apply LE_M1M1'.
  apply LE_lvlv'.

(* LURefRead *)
rename H into IH.

unfold NI_LRefReadEval_Prop.
intros LUR' Γ' Σ1' lv' Σ2' _t LE_M1M1' LE_LURLUR' EV'.

inversion LE_LURLUR' as [ | _mρ mρ' _UR UR' γ' LE_URUR' LE_mρmρ' ].  subst.

inversion EV' as [ | _UR' γ' ctx' _mρ' _Γ' _Σ1' _Σ2' uv' _lv' u' r' ]. subst.

assert (LE_M2M2'_uvuv' : LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\ exists l, LowEq_UValue l uv uv').
eapply IH.
  trivial.
  apply LE_M1M1'.
  apply LE_URUR'.
  apply u'.

destruct LE_M2M2'_uvuv' as [LE_M2M2' [ul LE_uvuv']].

split.
  apply LE_M2M2'.

  assert (LE_M2M2'_cl : LowEq_Memory (MkStack (lstack Γ) (ustack Γ) mρ,  Σ2) (MkStack (lstack Γ') (ustack Γ') mρ', Σ2')).
  apply LowEq_MClosure_Intro.
  apply LE_mρmρ'.
  apply LE_M2M2'.
  
  eapply UnlabeledCorrectnessOfLibraryModels.
  apply LE_M2M2'_cl.
  apply LE_uvuv'.
  apply r.
  apply r'.

(* UURefRead *)
unfold NI_URefReadEval_Prop.
intros UR' Γ' Σ1' uv' Σ2' _t LE_M1M1' LE_URUR' EV'.

inversion LE_URUR' as [ _p | ]. subst.

inversion EV' as [  ctx' _Γ' Σ' p' _uv' e' | ]. subst Σ1' Σ2'.  subst.

split.
  apply LE_M1M1'.
  exists H. constructor.

(* ULRefRead *)
rename H into IH.

unfold NI_URefReadEval_Prop.
intros UR' Γ' Σ1' uv' Σ3' _t LE_M1M1' LE_URUR' _EV'.

inversion LE_URUR' as [ _p | _LR LR' π' LE_LRLR' ]. subst.

inversion _EV' as [ | l' _LR' ctx' _Γ' _Σ1' π' ϕ' Σ2' lv' _Σ3' _uv' l0' EV' u']. subst.

assert (LE_M2M2'_lvlv' : LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\ LowEq_LValue lv lv').
eapply IH.
  trivial.
  apply LE_M1M1'.
  apply LE_LRLR'.
  apply EV'.

assert (EQ: Some (MVRefRead ϕ) = Some (MVRefRead ϕ')).
eapply NI_LookupM.
  apply LE_M1M1'.  
  apply l0.
  apply l0'.

injection EQ. intro. subst ϕ'.

assert (LE_M3M3'_uvuv' : LowEq_Memory (Γ, Σ3) (Γ', Σ3') /\ (exists l : Label, LowEq_UValue l uv uv')).
eapply NI_Unlabel_LValue.
  apply LE_M2M2'_lvlv'.
  apply LE_M2M2'_lvlv'.
  apply u.
  apply u'.

apply LE_M3M3'_uvuv'.
Qed.

Lemma NI_LRefReadEval :
  forall LR ctx Γ Σ1 Σ2 lv (EV : LRefReadEval LR ctx Γ Σ1 Σ2 lv), NI_LRefReadEval_Prop EV.
Proof.
apply NI_RefReadEval.
Qed. 

Lemma NI_URefReadEval :
  forall UR ctx Γ Σ1 Σ2 uv (EV : URefReadEval UR ctx Γ Σ1 Σ2 uv), NI_URefReadEval_Prop EV.
Proof. 
apply NI_RefReadEval.
Qed.

(* ******************************************************************************** *)


Definition NI_LRefWriteEval_Prop LW ctx Γ Σ1 lv Σ2 (EV : LRefWriteEval LW ctx Γ Σ1 lv Σ2) :=
  forall LW' Γ' Σ1' lv' Σ2',
    ctx = L ->
    LowEq_Memory (Γ, Σ1) (Γ', Σ1') ->
    LowEq_LRefWrite LW LW' ->
    LowEq_LValue lv lv' ->
    LRefWriteEval LW' ctx Γ' Σ1' lv' Σ2' ->
  LowEq_Memory (Γ, Σ2) (Γ', Σ2').

Definition NI_URefWriteEval_Prop UW ctx Γ Σ1 uv Σ2 (EV : URefWriteEval UW ctx Γ Σ1 uv Σ2) :=
  forall UW' Γ' Σ1' l uv' Σ2',
    ctx = L ->
    LowEq_Memory (Γ, Σ1) (Γ', Σ1') ->
    LowEq_URefWrite UW UW' ->
    LowEq_UValue l uv uv' ->
    URefWriteEval UW' ctx Γ' Σ1' uv' Σ2' ->
  LowEq_Memory (Γ, Σ2) (Γ', Σ2').

Lemma NI_RefWriteEval :
  (forall LW ctx Γ Σ1 lv Σ2 (EV : LRefWriteEval LW ctx Γ Σ1 lv Σ2), NI_LRefWriteEval_Prop EV) /\
  (forall UW ctx Γ Σ1 uv Σ2 (EV : URefWriteEval UW ctx Γ Σ1 uv Σ2), NI_URefWriteEval_Prop EV).
Proof.
apply RefWriteEval_mutind with
  (P  := NI_LRefWriteEval_Prop)
  (P0 := NI_URefWriteEval_Prop); intros.

(* LLRefWrite *)
unfold NI_LRefWriteEval_Prop.
intros LW' Γ' Σ1' lv' Σ2' _t LE_M1M1' LE_LWLW' LE_vv' EV'. 

inversion LE_LWLW' as [ _p p' bij | ]. subst.

inversion EV' as [ ctx' _Γ' _Σ1' _Σ2' _p' oldlv' _lv' e' c' e0' | ]. subst.

assert (LE_M2M2' : 
  LowEq_Memory 
    (Γ, MkHeap (Update (lheap Σ1) p (LIValue (Upgrade L lv))) (uheap Σ1) (mheap Σ1))
    (Γ', MkHeap (Update (lheap Σ1') p' (LIValue (Upgrade L lv'))) (uheap Σ1') (mheap Σ1'))
).

apply Bijection_LHeap_Update.
  rewrite MkHeap_Equal, MkHeap_Equal. apply LE_M1M1'.
  rewrite UpgradePublic, UpgradePublic. apply LE_vv'.
  apply bij.

apply LE_M2M2'.

(* LURefWrite *)
rename H into IH.

unfold NI_LRefWriteEval_Prop.
intros LUW' Γ' Σ1' lv' Σ3' _t LE_M1M1' LE_LUWLUW' LE_vv' _EV'. 

inversion LE_LUWLUW' as [ | _mρ mρ' _UW UW' γ' ϕ' LE_UWUW' LE_mρmρ' ]. subst.

inversion _EV' as [ | _mρ' _UW γ' ϕ' ctx' _Γ' _Σ1' oldl' _lv' uv' Σ2' _Σ3' i' c' u' u0' ]. subst.

assert (LE_M1M1'_cl : LowEq_Memory (MkStack (lstack Γ) (ustack Γ) mρ, Σ1) (MkStack (lstack Γ') (ustack Γ') mρ', Σ1')).
apply LowEq_MClosure_Intro.
  apply LE_mρmρ'.
  apply LE_M1M1'.

assert (LE_M2M2'_cl_uvuv' : 
  LowEq_Memory (MkStack (lstack Γ) (ustack Γ) mρ, Σ2) (MkStack (lstack Γ') (ustack Γ') mρ', Σ2') /\ 
  exists l, LowEq_UValue l uv uv').
eapply NI_Unlabel_LValue.
  apply LE_M1M1'_cl.
  apply LE_vv'.
  rewrite UpgradePublic in u.
  apply u.
  rewrite UpgradePublic in u'.
  apply u'.

destruct LE_M2M2'_cl_uvuv' as [LE_M2M2'_cl [ul LE_uvuv']].

assert (LE_M2M2' : LowEq_Memory (Γ, Σ2) (Γ', Σ2')).
eapply LowEq_MClosure_Elim.
  apply LE_M1M1'.
  apply LE_M2M2'_cl.

assert (LE_M3M3' : LowEq_Memory (Γ, Σ3) (Γ', Σ3')). 
eapply IH.
  trivial.
  apply LE_M2M2'.
  apply LE_UWUW'.
  apply LE_uvuv'.
  apply u0'.

apply LE_M3M3'.

(* UURefWrite *)
unfold NI_URefWriteEval_Prop.
intros UW' Γ' Σ1' l uv' Σ2' _t LE_M1M1' LE_UWUW' LE_uvuv' EV'.

inversion LE_UWUW' as [ p' | ]. subst. 

inversion EV' as [ ctx' _Γ' _Σ1' _Σ2' p' _uv' _t | ]. subst.

remember (MkHeap (lheap Σ1) (Update (uheap Σ1) p (UIValue uv)) (mheap Σ1)) as Σ2.
remember (MkHeap (lheap Σ1') (Update (uheap Σ1') p (UIValue uv')) (mheap Σ1')) as Σ2'.

(* Does not touch the part of the memory that is used by LowEq *)
inversion LE_M1M1'. subst.
constructor; assumption.

(* ULRefWrite *)
rename H into IH.

unfold NI_URefWriteEval_Prop.
intros UW' Γ' Σ1' ul uv' Σ2' _t LE_M1M1' LE_UWUW' LE_uvuv' _EV'.

inversion LE_UWUW' as [ | _LW LW' π' LE_LWLW' ]. subst.

inversion _EV' as [ | ctx' l1' l2' _LW' π' κ' γ' _Γ' _Σ1' _Σ2' _uv' lv' l' r' i' EV']. subst.

assert (EQ : Some (MVRefWrite κ γ) = Some (MVRefWrite κ' γ')).
eapply NI_LookupM.
  apply LE_M1M1'.
  apply l.
  apply l'.

injection EQ. intros. subst κ' γ'. clear EQ.

assert (LE_lvlv' : LowEq_LValue lv lv').
eapply UnlabeledCorrectnessOfLibraryModels.
  apply LE_M1M1'.
  apply LE_uvuv'.
  apply r.
  apply r'.
  
assert (EQ: l2 = l2').
eapply NI_Interp_LTerm.
  apply LE_M1M1'.
  apply i.
  apply i'.
subst l2'.

destruct l2.

(* l2 = l2' = L *)
assert (LE_M2M2' : LowEq_Memory (Γ, Σ2) (Γ', Σ2')).
eapply IH.
  trivial.
  apply LE_M1M1'.
  apply LE_LWLW'.
  apply LE_lvlv'.  
  apply EV'.

apply LE_M2M2'.

(* l2 = l2' = H *)

assert (LE_M1M2 : LowEq_Memory (Γ, Σ1) (Γ, Σ2)).
eapply Conf_LRefWriteEval.
  apply l0.
  trivial.
  eapply LowEq_Memory_Refl.
  apply LE_M1M1'.

assert (LE_M1'M2' : LowEq_Memory (Γ', Σ1') (Γ', Σ2')).
eapply Conf_LRefWriteEval.
  apply EV'.
  trivial.
  eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M1M1'.

assert (LE_M2M2' : LowEq_Memory (Γ, Σ2) (Γ', Σ2')).
eapply LowEq_Memory_Trans.
  apply LowEq_Memory_Symm.
  apply LE_M1M2.
  eapply LowEq_Memory_Trans.
  apply LE_M1M1'.
  apply LE_M1'M2'.

apply LE_M2M2'.
Qed.

Lemma NI_LRefWriteEval :
  forall LW ctx Γ Σ1 lv Σ2 (EV : LRefWriteEval LW ctx Γ Σ1 lv Σ2), NI_LRefWriteEval_Prop EV.
Proof.
apply NI_RefWriteEval.
Qed.

Lemma NI_URefWriteEval :
  forall UW ctx Γ Σ1 uv Σ2 (EV : URefWriteEval UW ctx Γ Σ1 uv Σ2), NI_URefWriteEval_Prop EV.
Proof.
apply NI_RefWriteEval.
Qed.

(* ******************************************************************************** *)

Definition NI_LHeadEval_Prop LHd ctx Γ Σ1 Σ2 lv (EV : LHeadEval LHd ctx Γ Σ1 Σ2 lv) :=
  forall LHd' Γ' Σ1' lv' Σ2',
  ctx = L ->
    LowEq_Memory (Γ, Σ1) (Γ', Σ1') ->
    LowEq_LHead LHd LHd' ->
    LHeadEval LHd' ctx Γ' Σ1' Σ2' lv' ->
  LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\
  LowEq_LValue lv lv'.


Definition NI_UHeadEval_Prop UHd ctx Γ Σ1 Σ2 uv (EV : UHeadEval UHd ctx Γ Σ1 Σ2 uv) :=
 forall UHd' Γ' Σ1' uv' Σ2',
  ctx = L ->
    LowEq_Memory (Γ, Σ1) (Γ', Σ1') ->
    LowEq_UHead UHd UHd' ->
    UHeadEval UHd' ctx Γ' Σ1' Σ2' uv' ->
  LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\
  exists l, LowEq_UValue l uv uv'.

Lemma NI_HeadEval :
  (forall LR ctx Γ Σ1 Σ2 lv (EV : LHeadEval LR ctx Γ Σ1 Σ2 lv), NI_LHeadEval_Prop EV) /\ 
  (forall UR ctx Γ Σ1 Σ2 uv (EV : UHeadEval UR ctx Γ Σ1 Σ2 uv), NI_UHeadEval_Prop EV).
Proof.  
apply HeadEval_mutind with
  (P  := NI_LHeadEval_Prop)
  (P0 := NI_UHeadEval_Prop); intros.

(* LLHead *)
unfold NI_LHeadEval_Prop.
intros LHd' Γ' Σ1' _lv' Σ2' _t LE_M1M1' LE_LHdLHd' EV'.

inversion LE_LHdLHd' as [_lv lv' LE_lvlv' | ]. subst.
inversion EV' as [ ctx' _Γ' Σ' __lv' | ]. subst _lv' __lv' Σ1' Σ2'. subst.

split.
  apply LE_M1M1'.
  apply LE_lvlv'.

(* LUHead *)
rename H into IH.
unfold NI_LHeadEval_Prop.

intros LHd' Γ' Σ1' lv' Σ2' _t LE_M1M1' LE_LHdLHd' EV'.
inversion LE_LHdLHd' as [ | _mρ mρ' _UHd UHd' _γ LE_UHdUHd' LE_mm']. subst.
inversion EV' as [ | ctx' _Γ' _Σ1' _Σ2' _UHd' _mρ' uv' γ' _lv' u' r']. subst.

assert (LE_M2M2'_uvuv' : LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\ (exists l : Label, LowEq_UValue l uv uv')).
eapply IH.
  trivial.
  apply LE_M1M1'.
  apply LE_UHdUHd'.
  apply u'.

destruct LE_M2M2'_uvuv' as [ LE_M2M2' [ul LE_uvuv']].

 assert (LE_M2M2'_cl : LowEq_Memory (MkStack (lstack Γ) (ustack Γ) mρ,  Σ2) (MkStack (lstack Γ') (ustack Γ') mρ', Σ2')).
  apply LowEq_MClosure_Intro.
  apply LE_mm'.
  apply LE_M2M2'.

assert (LE_lvlv' : LowEq_LValue lv lv').
eapply UnlabeledCorrectnessOfLibraryModels.
  apply LE_M2M2'_cl.  
  apply LE_uvuv'.
  apply r.
  apply r'. 

split.
  apply LE_M2M2'.
  apply LE_lvlv'.

(* UUHEad *)
rename H into IH.

unfold NI_UHeadEval_Prop.
intros UR' Γ' Σ1' uv' Σ3' _t LE_M1M1' LE_UHdUHd' EV'.

inversion LE_UHdUHd' as [ _mρ mρ' _LHd LHd' _ϕ LE_LHdLHd' LE_mm' | ]. subst.
inversion EV' as [ ctx' _Γ' _Σ1' Σ2' _Σ3' l' _LHd' _mρ' lv' ϕ' _uv' l0' u' | ]. subst.

assert (LE_M2M2'_lvlv' : LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\ LowEq_LValue lv lv').
eapply IH.
  trivial.
  apply LE_M1M1'.
  apply LE_LHdLHd'.
  apply l0'.

remember (MkStack (lstack Γ) (ustack Γ) mρ) as Γ2.
remember (MkStack (lstack Γ') (ustack Γ') mρ') as Γ2'.

assert (LE_M2M2'_cl : LowEq_Memory (Γ2, Σ2) (Γ2', Σ2')). subst.
eapply LowEq_MClosure_Intro.
  apply LE_mm'.
  apply LE_M2M2'_lvlv'.

assert (LE_M3M3'_cl_uvuv' : LowEq_Memory (Γ2, Σ3) (Γ2', Σ3') /\ (exists l : Label, LowEq_UValue l uv uv')). subst.
eapply NI_Unlabel_LValue.
  apply LE_M2M2'_cl.
  apply LE_M2M2'_lvlv'.
  rewrite UpgradePublic in u. apply u.
  rewrite UpgradePublic in u'. apply u'.

assert (LE_M3M3' : LowEq_Memory (Γ, Σ3) (Γ', Σ3')). subst.
eapply LowEq_MClosure_Elim.
  apply LE_M2M2'_lvlv'.
  apply LE_M3M3'_cl_uvuv'.

split.
  apply LE_M3M3'.
  apply LE_M3M3'_cl_uvuv'.

(* ULHead *)
unfold NI_UHeadEval_Prop.
intros UR' Γ' Σ1' _uv' Σ3' _t LE_M1M1' LE_UHdUHd' EV'.

inversion LE_UHdUHd' as [ | _uv uv' LE_uvuv']. subst.
inversion EV'. subst.

split.
  apply LE_M1M1'.
  exists L.
  apply LE_uvuv'.
Qed.

Lemma NI_LHeadEval :
  (forall LR ctx Γ Σ1 Σ2 lv (EV : LHeadEval LR ctx Γ Σ1 Σ2 lv), NI_LHeadEval_Prop EV).
Proof.
apply NI_HeadEval.
Qed.

Lemma NI_UHeadEval :
  (forall UR ctx Γ Σ1 Σ2 uv (EV : UHeadEval UR ctx Γ Σ1 Σ2 uv), NI_UHeadEval_Prop EV).
Proof.
apply NI_HeadEval.
Qed.

(* ******************************************************************************** *)

Definition NI_LTailEval_Prop LTl ctx Γ Σ1 Σ2 lv (EV : LTailEval LTl ctx Γ Σ1 Σ2 lv) :=
  forall LTl' Γ' Σ1' lv' Σ2',
  ctx = L ->
    LowEq_Memory (Γ, Σ1) (Γ', Σ1') ->
    LowEq_LTail LTl LTl' ->
    LTailEval LTl' ctx Γ' Σ1' Σ2' lv' ->
  LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\
  LowEq_LValue lv lv'.


Definition NI_UTailEval_Prop UTl ctx Γ Σ1 Σ2 uv (EV : UTailEval UTl ctx Γ Σ1 Σ2 uv) :=
 forall UTl' Γ' Σ1' uv' Σ2',
  ctx = L ->
    LowEq_Memory (Γ, Σ1) (Γ', Σ1') ->
    LowEq_UTail UTl UTl' ->
    UTailEval UTl' ctx Γ' Σ1' Σ2' uv' ->
  LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\
  exists l, LowEq_UValue l uv uv'.

Lemma NI_TailEval :
  (forall LTl ctx Γ Σ1 Σ2 lv (EV : LTailEval LTl ctx Γ Σ1 Σ2 lv), NI_LTailEval_Prop EV) /\ 
  (forall UTl ctx Γ Σ1 Σ2 uv (EV : UTailEval UTl ctx Γ Σ1 Σ2 uv), NI_UTailEval_Prop EV).
Proof.  
apply TailEval_mutind with
  (P  := NI_LTailEval_Prop)
  (P0 := NI_UTailEval_Prop); intros.

(* LLTail *)
unfold NI_LTailEval_Prop.
intros LHd' Γ' Σ1' _lv' Σ2' _t LE_M1M1' LE_LTlLTl' EV'.

inversion LE_LTlLTl' as [_lv lv' LE_lvlv' | ]. subst.
inversion EV' as [ ctx' _Γ' Σ' __lv' | ]. subst _lv' __lv' Σ1' Σ2'. subst.

split.
  apply LE_M1M1'.
  apply LE_lvlv'.

(* LUTail *)
rename H into IH.
unfold NI_LTailEval_Prop.

intros LHd' Γ' Σ1' lv' Σ2' _t LE_M1M1' LE_LTlLTl' EV'.
inversion LE_LTlLTl' as [ | _mρ mρ' _UTl UTl' _γ LE_UTlUTl' LE_mm']. subst.
inversion EV' as [ | ctx' _Γ' _Σ1' _Σ2' _UTl' _mρ' uv' γ' _lv' u' r']. subst.

assert (LE_M2M2'_uvuv' : LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\ (exists l : Label, LowEq_UValue l uv uv')).
eapply IH.
  trivial.
  apply LE_M1M1'.
  apply LE_UTlUTl'.
  apply u'.

destruct LE_M2M2'_uvuv' as [ LE_M2M2' [ul LE_uvuv']].

 assert (LE_M2M2'_cl : LowEq_Memory (MkStack (lstack Γ) (ustack Γ) mρ,  Σ2) (MkStack (lstack Γ') (ustack Γ') mρ', Σ2')).
  apply LowEq_MClosure_Intro.
  apply LE_mm'.
  apply LE_M2M2'.

assert (LE_lvlv' : LowEq_LValue lv lv').
eapply UnlabeledCorrectnessOfLibraryModels.
  apply LE_M2M2'_cl.  
  apply LE_uvuv'.
  apply r.
  apply r'. 

split.
  apply LE_M2M2'.
  apply LE_lvlv'.

(* UUTail *)
rename H into IH.

unfold NI_UTailEval_Prop.
intros UR' Γ' Σ1' uv' Σ3' _t LE_M1M1' LE_UTlUTl' EV'.

inversion LE_UTlUTl' as [ _mρ mρ' _LTl LTl' _ϕ LE_LTlLTl' LE_mm' | ]. subst.
inversion EV' as [ ctx' _Γ' _Σ1' Σ2' _Σ3' l' _LTl' _mρ' lv' ϕ' _uv' l0' u' | ]. subst.

assert (LE_M2M2'_lvlv' : LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\ LowEq_LValue lv lv').
eapply IH.
  trivial.
  apply LE_M1M1'.
  apply LE_LTlLTl'.
  apply l0'.

remember (MkStack (lstack Γ) (ustack Γ) mρ) as Γ2.
remember (MkStack (lstack Γ') (ustack Γ') mρ') as Γ2'.

assert (LE_M2M2'_cl : LowEq_Memory (Γ2, Σ2) (Γ2', Σ2')). subst.
eapply LowEq_MClosure_Intro.
  apply LE_mm'.
  apply LE_M2M2'_lvlv'.

assert (LE_M3M3'_cl_uvuv' : LowEq_Memory (Γ2, Σ3) (Γ2', Σ3') /\ (exists l : Label, LowEq_UValue l uv uv')). subst.
eapply NI_Unlabel_LValue.
  apply LE_M2M2'_cl.
  apply LE_M2M2'_lvlv'.
  rewrite UpgradePublic in u. apply u.
  rewrite UpgradePublic in u'. apply u'.

assert (LE_M3M3' : LowEq_Memory (Γ, Σ3) (Γ', Σ3')). subst.
eapply LowEq_MClosure_Elim.
  apply LE_M2M2'_lvlv'.
  apply LE_M3M3'_cl_uvuv'.

split.
  apply LE_M3M3'.
  apply LE_M3M3'_cl_uvuv'.

(* ULTail *)
unfold NI_UTailEval_Prop.
intros UR' Γ' Σ1' _uv' Σ3' _t LE_M1M1' LE_UTlUTl' EV'.

inversion LE_UTlUTl' as [ | _uv uv' LE_uvuv']. subst.
inversion EV'. subst.

split.
  apply LE_M1M1'.
  exists L.
  apply LE_uvuv'.
Qed.

Lemma NI_LTailEval :
  (forall LTl ctx Γ Σ1 Σ2 lv (EV : LTailEval LTl ctx Γ Σ1 Σ2 lv), NI_LTailEval_Prop EV).
Proof.
apply NI_TailEval.
Qed.

Lemma NI_UTailEval :
  (forall UTl ctx Γ Σ1 Σ2 uv (EV : UTailEval UTl ctx Γ Σ1 Σ2 uv), NI_UTailEval_Prop EV).
Proof.
apply NI_TailEval.
Qed.

(* ******************************************************************************** *)

Definition NI_LFunEval_Prop F ctx Γ Σ1 lv1 Σ2 lv2 (EV : LFunEval F ctx Γ Σ1 lv1 Σ2 lv2) :=
  forall F' Γ' Σ1' lv1' Σ2' lv2',
    ctx = L ->
    LowEq_Memory (Γ, Σ1) (Γ', Σ1') ->
    LowEq_LFun F F' ->
    LowEq_LValue lv1 lv1' -> 
    LFunEval F' ctx Γ' Σ1' lv1' Σ2' lv2' ->
  LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\
  LowEq_LValue lv2 lv2'.

Definition NI_UFunEval_Prop F ctx Γ Σ1 uv1 Σ2 uv2 (EV : UFunEval F ctx Γ Σ1 uv1 Σ2 uv2) :=
  forall F' Γ' Σ1' uv1' Σ2' uv2' l,
    ctx = L ->
    LowEq_Memory (Γ, Σ1) (Γ', Σ1') ->
    LowEq_UFun L F F' ->
    LowEq_UValue l uv1 uv1' -> 
    UFunEval F' ctx Γ' Σ1' uv1' Σ2' uv2' ->
  LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\
  exists l, LowEq_UValue l uv2 uv2'.

Definition NI_LEval_Prop ctx Γ Σ1 e Σ2 lv (EV : LEval ctx Γ Σ1 e Σ2 lv) :=
  forall Γ' Σ1' Σ2' lv',
    ctx = L ->
    LowEq_Memory (Γ, Σ1) (Γ', Σ1') ->
    LEval L Γ' Σ1' e (* => *) Σ2' lv' ->
  LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\
  LowEq_LValue lv lv'.

Definition NI_UEval_Prop ctx Γ Σ1 e Σ2 uv (EV : UEval ctx Γ Σ1 e Σ2 uv) := True.

Theorem Noninterference :
  (forall F ctx Γ Σ1 lv1 Σ2 lv2 (EV : LFunEval F ctx Γ Σ1 lv1 Σ2 lv2), NI_LFunEval_Prop EV) /\
  (forall F ctx Γ Σ1 uv1 Σ2 uv2 (EV : UFunEval F ctx Γ Σ1 uv1 Σ2 uv2), NI_UFunEval_Prop EV) /\ 
  (forall ctx Γ Σ1 e Σ2 lv (EV : LEval ctx Γ Σ1 e Σ2 lv), NI_LEval_Prop EV) /\
  (forall ctx Γ Σ1 e Σ2 uv (EV : UEval ctx Γ Σ1 e Σ2 uv), NI_UEval_Prop EV).
Proof.
apply Eval_mutind with
  (P  := NI_LFunEval_Prop)
  (P0 := NI_UFunEval_Prop)
  (P1 := NI_LEval_Prop)
  (P2 := NI_UEval_Prop); intros.

(* ******************************************************************************** *)
(* LLFun *)
rename H into IH.

unfold NI_LFunEval_Prop.
intros F' Γ' Σ1' lv1' Σ3' lv2' _t LE_M1M1' LE_FF' LE_lv1lv1' _EV'.
inversion LE_FF'. subst.
inversion _EV' as [ ctx' _Γ _Σ1' Σ2' _Σ3' p' σl' ρl' x' e' vp' vr' f' e0' e1' EV' | ].

(* LLFun, LLFun *)
subst lρ2. subst.
rename vr into lv2. rename vp into lv1.

remember (MkHeap (Update (lheap Σ1) p (LIFrame (Singleton x (Some (Upgrade L lv1))))) (uheap Σ1) (mheap Σ1)) as Σ2.
remember (MkHeap (Update (lheap Σ1') p' (LIFrame (Singleton x (Some (Upgrade L lv1'))))) (uheap Σ1') (mheap Σ1')) as Σ2'.
remember (MkStack (p :: ρl) (ustack Γ) (mstack Γ)) as Γ2.
remember (MkStack (p' :: ρl') (ustack Γ') (mstack Γ')) as Γ2'.

assert (LE_M1M1'_cl : 
  LowEq_Memory 
    (MkStack ρl (ustack Γ) (mstack Γ), MkHeap (lheap Σ1) (uheap Σ1) (mheap Σ1))
    (MkStack ρl' (ustack Γ') (mstack Γ'), MkHeap (lheap Σ1') (uheap Σ1') (mheap Σ1'))
).
apply LowEq_LClosure_Intro.
  assumption. (* LowEq_LStack ρl ρl' *)
  rewrite MkHeap_Equal, MkHeap_Equal. apply LE_M1M1'.

assert (bij : extension bl p p').
eapply Fresh_LBijection.
  apply f.  
  apply f'.

assert (LE_M2M2'_cl_ext : LowEq_Memory (Γ2, Σ2) (Γ2', Σ2')). subst.
apply Bijection_Memory_LUpdate.
  apply LE_M1M1'_cl.
  constructor.
  apply LowEq_Singleton_LFrame.
  rewrite UpgradePublic, UpgradePublic. apply LE_lv1lv1'.
  apply bij.

assert (LE_M3M3'_cl_ext_lv2lv2': LowEq_Memory (Γ2, Σ3) (Γ2', Σ3') /\
                   LowEq_LValue lv2 lv2').
unfold NI_LEval_Prop in IH.
eapply IH.
  trivial.
  apply LE_M2M2'_cl_ext.
  apply EV'.

destruct LE_M3M3'_cl_ext_lv2lv2' as [LE_M3M3'_cl_ext LE_lv2lv2']. 

assert (LE_M2M2' : LowEq_Memory (Γ, Σ2) (Γ', Σ2')). subst.
eapply LowEq_LClosure_Elim.
  apply LE_M1M1'.
  apply LE_M2M2'_cl_ext.

assert (LE_M3M3' : LowEq_Memory (Γ, Σ3) (Γ', Σ3')). subst.
eapply LowEq_LClosure_Elim.
  apply LE_M2M2'.
  apply LE_M3M3'_cl_ext.

constructor.
  apply LE_M3M3'.
  apply LE_lv2lv2'.

(* ******************************************************************************** *)
(* LUFun *)
rename H into IH.

unfold NI_LFunEval_Prop.
intros F' Γ' Σ1' lv1' Σ3' lv2' _t LE_M1M1' LE_FF' LE_lv1lv1' _EV'.
inversion LE_FF' as [ | _mρ mρ' _UF UF' _γ _ϕ _ξ LE_UFUF' LE_mρmρ' ] . subst.
rename Σ3' into Σ5'. rename Γ' into Γ1'.
inversion _EV' as [ | _mρ _UF' _ϕ _γ' _ξ ctx p' _Γ1' Γ2' _Σ1' Σ2' Σ3' Σ4' _Σ5' _lv1' _lv2' uv1' uv2' f' _t1 _t2 i' u'  u0' r'].
subst.

remember (MkHeap (lheap Σ1) (uheap Σ1) (Update (mheap Σ1) p (MIFrame EmptyMFrame))) as Σ2.
remember (MkHeap (lheap Σ1') (uheap Σ1') (Update (mheap Σ1') p' (MIFrame EmptyMFrame))) as Σ2'.
remember (MkStack (lstack Γ1) (ustack Γ1) (p :: mρ)) as Γ2.
remember (MkStack (lstack Γ1') (ustack Γ1') (p' :: mρ')) as Γ2'.

assert (LE_M1M1'_cl :
  LowEq_Memory 
    (MkStack (lstack Γ1) (ustack Γ1) mρ, Σ1) 
    (MkStack (lstack Γ1') (ustack Γ1') mρ', Σ1')
).
apply LowEq_MClosure_Intro.
  apply LE_mρmρ'.
  apply LE_M1M1'.

assert (bij : extension bm p p').
eapply Fresh_MBijection.
  apply f0.
  apply f'.

assert (LE_M2M2'_cl_ext : LowEq_Memory (Γ2, Σ2) (Γ2', Σ2')). subst.
apply Bijection_Memory_MUpdate.
  rewrite MkHeap_Equal, MkHeap_Equal.
  apply LE_M1M1'_cl.
  constructor.
    apply LowEq_EmptyMFrame.  
  apply bij.
 
assert (LE_M3M3'_cl_ext_uv1uv1' : 
  LowEq_Memory (Γ2, Σ3) (Γ2', Σ3') /\ 
  exists l, LowEq_UValue l uv1 uv1'
).
eapply NI_Unlabel_LValue.
  apply LE_M2M2'_cl_ext.
  apply LE_lv1lv1'.
  apply u.
  apply i'.

destruct LE_M3M3'_cl_ext_uv1uv1' as [LE_M3M3'_cl_ext [ul1 LE_uv1uv1']].

assert (LE_M4M4'_cl_ext : LowEq_Memory (Γ2, Σ4) (Γ2', Σ4')).
eapply NI_Interp_EffCs.
  apply LE_M3M3'_cl_ext.
  apply i.
  apply u'.

assert (LE_M2M2' : LowEq_Memory (Γ1, Σ2) (Γ1', Σ2')). subst.
eapply LowEq_MClosure_Elim.
  apply LE_M1M1'.
  apply LE_M2M2'_cl_ext.

   
assert (LE_M3M3' : LowEq_Memory (Γ1, Σ3) (Γ1', Σ3')). subst.
eapply LowEq_MClosure_Elim.
  apply LE_M2M2'.
  apply LE_M3M3'_cl_ext.

assert (LE_M4M4' : LowEq_Memory (Γ1, Σ4) (Γ1', Σ4')). subst.
eapply LowEq_MClosure_Elim.
  apply LE_M3M3'.
  apply LE_M4M4'_cl_ext.

assert (LE_M5M5'_uv2uv2' :  LowEq_Memory (Γ1, Σ5) (Γ1', Σ5') /\ exists l, LowEq_UValue l uv2 uv2').

eapply IH.
  trivial.
  apply LE_M4M4'.
  apply LE_UFUF'.
  apply LE_uv1uv1'.
  apply u0'.

destruct LE_M5M5'_uv2uv2' as [LE_M5M5' [ul LE_uv2uv2']].


 assert (LE_M5M5'_cl : LowEq_Memory (MkStack (lstack Γ1) (ustack Γ1) (p :: mρ),  Σ5) (MkStack (lstack Γ1') (ustack Γ1') (p' :: mρ'), Σ5')).
  apply LowEq_MClosure_Intro.
  constructor.
    eapply Fresh_MBijection.
    apply f0.
    apply f'.
  apply LE_mρmρ'.
  apply LE_M5M5'.

constructor.
  apply LE_M5M5'.
  subst.
  eapply UnlabeledCorrectnessOfLibraryModels.
    apply LE_M5M5'_cl.
    apply LE_uv2uv2'.
    apply r.  
    apply r'.

(* ******************************************************************************** *)
(* UUFun *)
rename H into IH.

unfold NI_UFunEval_Prop.
rename vp into uv1. rename vr into uv2. rename Γ into Γ1. 
intros F' Γ1' Σ1' uv1' Σ3' uv2' l _t LE_M1M1' LE_FF' LE_uv1uv1 _EV'.  
inversion LE_FF' as [ _ρu ρu' _x _e | ].
subst.
inversion _EV' as [ ctx' _Γ1' _Σ1' Σ2' _Σ3' p' σu' _ρu' x' e' _uv1' _uv2' f' e0' e1' u'| ]. subst.

remember (MkHeap (lheap Σ1) (Update (uheap Σ1) p (UIFrame (Singleton x (Some uv1)))) (mheap Σ1)) as Σ2. 
remember (MkHeap (lheap Σ1') (Update (uheap Σ1') p' (UIFrame (Singleton x (Some uv1')))) (mheap Σ1')) as Σ2'.
remember (MkStack (lstack Γ1) (p :: ρu) (mstack Γ1)) as Γ2.
remember (MkStack (lstack Γ1') (p' :: ρu') (mstack Γ1')) as Γ2'.

assert (LE_M2M2'_cl_ext : LowEq_Memory (Γ2, Σ2) (Γ2', Σ2')). subst.
eapply LowEq_UMemory.
  rewrite MkHeap_Equal, MkHeap_Equal, MkStack_Equal, MkStack_Equal.
  apply LE_M1M1'.

assert (LE_M3M3'_cl_ext : LowEq_Memory (Γ2, Σ3) (Γ2', Σ3')).
eapply LabeledCorrectnessOfLibraryModels.
  apply LE_M2M2'_cl_ext.
  apply u.
  apply u'.
  
assert (ex_LE_uv2uv2': exists l0 : Label, LowEq_UValue l0 uv2 uv2').
eapply LabelabilityOfUnlabeledExecution.
  apply LE_M2M2'_cl_ext.
  apply u.
  apply u'.

assert (LE_M2M2' : LowEq_Memory (Γ1, Σ2) (Γ1', Σ2')). subst.
eapply LowEq_Memory_UHeap_Mod.
 rewrite MkHeap_Equal, MkHeap_Equal.
 apply LE_M1M1'.

assert (LE_M3M3' : LowEq_Memory (Γ1, Σ3) (Γ1', Σ3')). subst.
destruct Γ1 as [lρ uρ mρ]. destruct Γ1' as [lρ' uρ' mρ'].
eapply LowEq_Memory_UStack_Mod.
  apply LE_M3M3'_cl_ext.

constructor.
  apply LE_M3M3'.
  apply ex_LE_uv2uv2'.

(* ******************************************************************************** *)
(* ULFun *)
rename H into IH.

unfold NI_UFunEval_Prop.
rename Γ into Γ1. 
intros F' Γ1' Σ1' uv1' Σ3' uv2' ul _t LE_M1M1' LE_FF' LE_uv1uv1' _EV'. subst.
inversion LE_FF' as [ | _LF LF' _π LE_LFLF']. subst.
inversion _EV' as [ | l1' l2' _LF' π' ctx' Γ' _Σ1' Σ2' _Σ3' κ' γ' ϕ' _uv1' _uv2' lv1' lv2' l' r' i' l0' u']. subst.

assert (EQ :  Some (MVApp κ γ ϕ) = Some (MVApp κ' γ' ϕ')).
eapply NI_LookupM.
  apply LE_M1M1'.
  apply l.
  apply l'.

injection EQ. intros. subst κ' γ' ϕ. 
clear EQ.

assert (LE_lv1lv1' : LowEq_LValue lv1 lv1'). 
eapply UnlabeledCorrectnessOfLibraryModels.
  apply LE_M1M1'.
  apply LE_uv1uv1'. 
  apply r.
  apply r'.

assert (EQ : l2 = l2').
eapply NI_Interp_LTerm.
  apply LE_M1M1'.
  apply i.  
  apply i'.
subst l2'.

destruct l2.

(* Context is L *)
assert (LE_M2M2'_lv2lv2' : LowEq_Memory (Γ1, Σ2) (Γ1', Σ2') /\ LowEq_LValue lv2 lv2').
eapply IH.
  trivial.
  apply LE_M1M1'.
  apply LE_LFLF'.
  apply LE_lv1lv1'.
  apply l0'.

destruct LE_M2M2'_lv2lv2' as [LE_M2M2' LE_lv2lv2'].

assert (LE_M3M3'_uv2uv2': LowEq_Memory (Γ1, Σ3) (Γ1', Σ3') /\ (exists l : Label, LowEq_UValue l uv2 uv2')).
eapply NI_Unlabel_LValue.
 apply LE_M2M2'.
 apply LE_lv2lv2'.
 rewrite UpgradePublic in u.
 apply u.
 rewrite UpgradePublic in u'.
 apply u'.
 
apply LE_M3M3'_uv2uv2'.

(* Context is H *)

assert (LE_M1M2 : LowEq_Memory (Γ1, Σ1) (Γ1, Σ2)).
eapply Conf_LFunEval.
  apply l0.
  auto.
  eapply LowEq_Memory_Refl.
  apply LE_M1M1'.

assert (LE_M1'M2' : LowEq_Memory (Γ1', Σ1') (Γ1', Σ2')).
eapply Conf_LFunEval.
  apply l0'.
  auto.
  eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M1M1'.

assert (LE_M2M2' : LowEq_Memory (Γ1, Σ2) (Γ1', Σ2')).
eapply LowEq_Memory_Trans.
  apply LowEq_Memory_Symm.
  apply LE_M1M2.
  eapply LowEq_Memory_Trans.
  apply LE_M1M1'.
  apply LE_M1'M2'.

assert (LE_M2M3 : LowEq_Memory (Γ1, Σ2) (Γ1, Σ3)).
eapply Conf_Unlabel_LValue.
  apply u.
  auto.
  eapply LowEq_Memory_Refl.
  apply LE_M2M2'.

assert (LE_M2'M3' : LowEq_Memory (Γ1', Σ2') (Γ1', Σ3')).
eapply Conf_Unlabel_LValue.
  apply u'.
  auto.
  eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M2M2'.

assert (LE_M3M3' : LowEq_Memory (Γ1, Σ3) (Γ1', Σ3')).
eapply LowEq_Memory_Trans.
  apply LowEq_Memory_Symm.
  apply LE_M2M3.
  eapply LowEq_Memory_Trans.
  apply LE_M2M2'.
  apply LE_M2'M3'.

constructor.
  apply LE_M3M3'.

assert (LE_lv2lv2': LowEq_LValue (Upgrade H lv2) (Upgrade H lv2')).
apply LowEq_Secret.

exists H. constructor.

(* ******************************************************************************** *)
(* Num *)
unfold NI_LEval_Prop.
intros Γ' Σ1' Σ2' lv' _t LE_M1M1' _EV'.
inversion _EV'. subst.
constructor.
  apply LE_M1M1'.
  constructor.

(* ******************************************************************************** *)
(* Var *)
unfold NI_LEval_Prop.
intros Γ' Σ1' Σ2' v' _t LE_M1M1' _EV'. subst.
inversion _EV' as [ | ctx _Γ' Σ' x' _v' l' | | | | | | | | | | | | | ]. subst Σ1' Σ2'. subst.

assert (LE_vv' : LowEq_LValue v v').
eapply NI_LookupL.
  apply LE_M1M1'.
  apply l.
  apply l'.

constructor.
  apply LE_M1M1'.
  apply LE_vv'.

(* ******************************************************************************** *)
(* If1 *)
rename H into IH1.
rename H0 into IH2.

unfold NI_LEval_Prop.
intros Γ' Σ1' Σ3' v' _t LE_M1M1' _EV'. subst.
inversion _EV' as [ | |  ctx _Γ' _Σ1' Σ2' _Σ3' e1' e2' e3' l' n' v2' EV1' n0' EV2' | 
                         ctx _Γ' _Σ1' Σ2' _Σ3' e1' e2' e3' l' n' v3' EV1' n0' EV2'| | | | | | | | | | | ]. 

(* If1, If1 *)
subst.
assert (LE_M2M2'_nn' : LowEq_Memory (Γ, Σ2) (Γ', Σ2')/\ LowEq_LValue (LVNum l n) (LVNum l' n')).
eapply IH1.
  trivial.
  apply LE_M1M1'.
  apply EV1'.

destruct LE_M2M2'_nn' as [LE_M2M2' LE_nn'].

inversion LE_nn' as [ | | | | |  _v1 _v2 LO LO' ].
subst.

(* l = l' = L *)

assert (LE_M3M3'_v2v2' : LowEq_Memory (Γ, Σ3) (Γ', Σ3')/\ LowEq_LValue v2 v2').
eapply IH2.
  trivial.
  apply LE_M2M2'.
  apply EV2'.

rewrite UpgradePublic.
rewrite UpgradePublic.

apply LE_M3M3'_v2v2'.

(* l = l' = H *)
unfold LabelOf in LO, LO'.
subst.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply Conf_LEval.
  apply l1.
  auto.
  eapply LowEq_Memory_Refl.
  apply LE_M2M2'.

assert (LE_M2'M3' : LowEq_Memory (Γ', Σ2') (Γ', Σ3')).
eapply Conf_LEval.
  apply EV2'.
  auto.
  eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M2M2'.

assert (LE_M3M3' : LowEq_Memory (Γ, Σ3) (Γ', Σ3')).
eapply LowEq_Memory_Trans.
  apply LowEq_Memory_Symm.
  apply LE_M2M3.
  eapply LowEq_Memory_Trans.
  apply LE_M2M2'.
  apply LE_M2'M3'.

constructor.
  apply LE_M3M3'.
  apply LowEq_Secret.

(* If1, If2 *)
subst.
assert (LE_M2M2'_nn' : LowEq_Memory (Γ, Σ2) (Γ', Σ2')/\ LowEq_LValue (LVNum l n) (LVNum l' (Num 0))).
eapply IH1.
  trivial.
  apply LE_M1M1'.
  apply EV1'.

destruct LE_M2M2'_nn' as [LE_M2M2' LE_nn'].

inversion LE_nn' as [ | | | | | _v1 _v2 LO LO' ].
subst.

(* l = l' = L *)
contradiction n0. trivial.

(* l = l' = H *)
unfold LabelOf in LO, LO'.
subst.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply Conf_LEval.
  apply l1.
  auto.
  eapply LowEq_Memory_Refl.
  apply LE_M2M2'.

assert (LE_M2'M3' : LowEq_Memory (Γ', Σ2') (Γ', Σ3')).
eapply Conf_LEval.
  apply EV2'.
  auto.
  eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M2M2'.

assert (LE_M3M3' : LowEq_Memory (Γ, Σ3) (Γ', Σ3')).
eapply LowEq_Memory_Trans.
  apply LowEq_Memory_Symm.
  apply LE_M2M3.
  eapply LowEq_Memory_Trans.
  apply LE_M2M2'.
  apply LE_M2'M3'.

constructor.
  apply LE_M3M3'.
  apply LowEq_Secret.

(* ******************************************************************************** *)
(* If2 *)
rename H into IH1.
rename H0 into IH2.

unfold NI_LEval_Prop.
intros Γ' Σ1' Σ3' v' _t LE_M1M1' _EV'. subst.
inversion _EV' as [ | |  ctx _Γ' _Σ1' Σ2' _Σ3' e1' e2' e3' l' n' v2' EV1' n0' EV2' | 
                         ctx _Γ' _Σ1' Σ2' _Σ3' e1' e2' e3' l' n' v3' EV1' n0' EV2'| | | | | | | | | | | ]. 


(* If2, If1 *)
subst.
assert (LE_M2M2'_nn' : LowEq_Memory (Γ, Σ2) (Γ', Σ2')/\ LowEq_LValue (LVNum l (Num 0)) (LVNum l' n')).
eapply IH1.
  trivial.
  apply LE_M1M1'.
  apply EV1'.

destruct LE_M2M2'_nn' as [LE_M2M2' LE_nn'].

inversion LE_nn' as [ | | | | | _v1 _v2 LO LO' ].
subst.

(* l = l' = L *)
contradiction n0'. trivial.

(* l = l' = H *)
unfold LabelOf in LO, LO'.
subst.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply Conf_LEval.
  apply l1.
  auto.
  eapply LowEq_Memory_Refl.
  apply LE_M2M2'.

assert (LE_M2'M3' : LowEq_Memory (Γ', Σ2') (Γ', Σ3')).
eapply Conf_LEval.
  apply EV2'.
  auto.
  eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M2M2'.

assert (LE_M3M3' : LowEq_Memory (Γ, Σ3) (Γ', Σ3')).
eapply LowEq_Memory_Trans.
  apply LowEq_Memory_Symm.
  apply LE_M2M3.
  eapply LowEq_Memory_Trans.
  apply LE_M2M2'.
  apply LE_M2'M3'.

constructor.
  apply LE_M3M3'.
  apply LowEq_Secret.


(* If2, If2 *)
subst.
assert (LE_M2M2'_nn' : LowEq_Memory (Γ, Σ2) (Γ', Σ2')/\ LowEq_LValue (LVNum l (Num 0)) (LVNum l' (Num 0))).
eapply IH1.
  trivial.
  apply LE_M1M1'.
  apply EV1'.

destruct LE_M2M2'_nn' as [LE_M2M2' LE_nn'].

inversion LE_nn' as [ | | | | | _v1 _v2 LO LO' ].
subst.

(* l = l' = L *)

assert (LE_M3M3'_v3v3' : LowEq_Memory (Γ, Σ3) (Γ', Σ3')/\ LowEq_LValue v3 v3').
eapply IH2.
  trivial.
  apply LE_M2M2'.
  apply EV2'.

rewrite UpgradePublic.
rewrite UpgradePublic.

apply LE_M3M3'_v3v3'.

(* l = l' = H *)
unfold LabelOf in LO, LO'.
subst.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply Conf_LEval.
  apply l1.
  auto.
  eapply LowEq_Memory_Refl.
  apply LE_M2M2'.

assert (LE_M2'M3' : LowEq_Memory (Γ', Σ2') (Γ', Σ3')).
eapply Conf_LEval.
  apply EV2'.
  auto.
  eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M2M2'.

assert (LE_M3M3' : LowEq_Memory (Γ, Σ3) (Γ', Σ3')).
eapply LowEq_Memory_Trans.
  apply LowEq_Memory_Symm.
  apply LE_M2M3.
  eapply LowEq_Memory_Trans.
  apply LE_M2M2'.
  apply LE_M2'M3'.

constructor.
  apply LE_M3M3'.
  apply LowEq_Secret.

(* ******************************************************************************** *)
(* Seq *)
rename H into IH1.
rename H0 into IH2.

unfold NI_LEval_Prop.
intros Γ' Σ1' Σ3' v2' _t LE_M1M1' _EV'. subst.
inversion _EV' as [ | | | |  ctx _Γ' _Σ1' Σ2' _Σ3' e1' e2' v1' _v2' EV1' EV2' | | | | | | | | | | ].
subst.

assert (LE_M2M2' : LowEq_Memory (Γ, Σ2) (Γ', Σ2')).
eapply IH1.
  trivial.
  apply LE_M1M1'.
  apply EV1'.

assert (LE_M3M3'_v2v2' : LowEq_Memory (Γ, Σ3) (Γ', Σ3') /\ LowEq_LValue v2 v2').
eapply IH2.
  trivial.
  apply LE_M2M2'.
  apply EV2'.

apply LE_M3M3'_v2v2'.

(* ******************************************************************************** *)
(* Ref *)
rename H into IH.

unfold NI_LEval_Prop.
intros Γ' Σ1' Σ3' v2' _t LE_M1M1' _EV'. subst.
inversion _EV' as [ | | | | | ctx _Γ' _Σ1' Σ2' _Σ3' e' v' LR' LW' EV' m' | | | | | | | | | ]. subst.

assert (LE_M2M2'_vv' : LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\ LowEq_LValue v v').
eapply IH.
  trivial.
  apply LE_M1M1'.
  apply EV'.

destruct LE_M2M2'_vv' as [LE_M2M2' LE_vv']. 

assert (LE_M3M3'_RR' : LowEq_Memory (Γ, Σ3) (Γ', Σ3') /\ LowEq_LValue (LVRef L LR LW) (LVRef L LR' LW')).
eapply NI_MkLRef.
  apply LE_M2M2'.
  apply LE_vv'.
  apply m.
  apply m'. 

apply LE_M3M3'_RR'.

(* ******************************************************************************** *)
(* Deref *)
rename H into IH.

unfold NI_LEval_Prop.
intros Γ' Σ1' Σ3' _v' _t LE_M1M1' _EV'. subst.
inversion _EV' as [ | | | | | | ctx _Γ' _Σ1' Σ2' _Σ3' e' l' LR' LW' v' EV' l0' | | | | | | | | ]. subst.

assert (LE_M2M2'_RR' : LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\ LowEq_LValue (LVRef l LR LW) (LVRef l' LR' LW')).
eapply IH.
  trivial.
  apply LE_M1M1'.
  apply EV'.

destruct LE_M2M2'_RR' as [LE_M2M2' LE_LRLR'].
inversion LE_LRLR' as [ | | R R' W W' LE_RR' LE_WW' | | | _v1 _v2 LO LO'].

(* l = l' = L *)
subst.

assert (LE_M3M3'_vv' : LowEq_Memory (Γ, Σ3) (Γ', Σ3') /\ LowEq_LValue v v').
eapply NI_LRefReadEval.
  apply l1.
  trivial.
  apply LE_M2M2'.
  apply LE_RR'.
  apply l0'.

rewrite UpgradePublic.
rewrite UpgradePublic.
apply LE_M3M3'_vv'.

(* l = l' = H *)
unfold LabelOf in LO, LO'.
subst.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply Conf_LRefReadEval.
  apply l1.
  auto.
  eapply LowEq_Memory_Refl.
  apply LE_M2M2'.

assert (LE_M2'M3' : LowEq_Memory (Γ', Σ2') (Γ', Σ3')).
eapply Conf_LRefReadEval.
  apply l0'.
  auto.
  eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M2M2'.

assert (LE_M3M3' : LowEq_Memory (Γ, Σ3) (Γ', Σ3')).
eapply LowEq_Memory_Trans.
  apply LowEq_Memory_Symm.
  apply LE_M2M3.
  eapply LowEq_Memory_Trans.
  apply LE_M2M2'.
  apply LE_M2'M3'.

constructor.
  apply LE_M3M3'.
  apply LowEq_Secret.

(* ******************************************************************************** *)
(* Asn *)
rename H into IH1.
rename H0 into IH2.

unfold NI_LEval_Prop.
intros Γ' Σ1' Σ4' v' _t LE_M1M1' _EV'. subst.
inversion _EV' as [ | | | | | | |  ctx _Γ' _Σ1' Σ2' Σ3' _Σ4' e1' e2' l' R' W' _v' EV1' EV2' l0' | | | | | | | ]. subst.

assert (LE_M2M2'_RR' : LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\ LowEq_LValue (LVRef l R W) (LVRef l' R' W')).
eapply IH1.
  trivial.
  apply LE_M1M1'.
  apply EV1'.

destruct LE_M2M2'_RR' as [LE_M2M2' LE_LLRLLR'].

assert (LE_M3M3'_vv' : LowEq_Memory (Γ, Σ3) (Γ', Σ3') /\ LowEq_LValue v v').
eapply IH2.
  trivial.
  apply LE_M2M2'.
  apply EV2'.

destruct LE_M3M3'_vv' as [LE_M3M3' LE_vv'].

inversion LE_LLRLLR' as [ | | _R _R' _W _W' LE_RR' LE_WW' | | | _v1 _v2 LO LO'].

(* l = l' = L *)
subst.

assert (LE_M4M4' : LowEq_Memory (Γ, Σ4) (Γ', Σ4')).
eapply NI_LRefWriteEval.
  apply l2.
  trivial.
  apply LE_M3M3'.
  apply LE_WW'.
  apply LE_vv'.
  apply l0'.

constructor.
  apply LE_M4M4'.
  apply LE_vv'.

(* l = l' = H *)

unfold LabelOf in LO, LO'.
subst.

assert (LE_M3M4 : LowEq_Memory (Γ, Σ3) (Γ, Σ4)).
eapply Conf_LRefWriteEval.
  apply l2.
  auto.
  eapply LowEq_Memory_Refl.
  apply LE_M3M3'.

assert (LE_M3'M4' : LowEq_Memory (Γ', Σ3') (Γ', Σ4')).
eapply Conf_LRefWriteEval.
  apply l0'.
  auto.
  eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M3M3'.

assert (LE_M4M4' : LowEq_Memory (Γ, Σ4) (Γ', Σ4')).
eapply LowEq_Memory_Trans.
  apply LowEq_Memory_Symm.
  apply LE_M3M4.
  eapply LowEq_Memory_Trans.
  apply LE_M3M3'.
  apply LE_M3'M4'.

constructor.
  apply LE_M4M4'.
  apply LE_vv'.

(* ******************************************************************************** *)
(* Fun *)
 
unfold NI_LEval_Prop.
intros Γ' Σ1' _Σ1' v' _t LE_M1M1' _EV'. subst.
inversion _EV' as [ | | | | | | | | ctx _Γ' Σ' x' e' F' e0' | | | | | | ]. subst Σ1' _Σ1'. subst.
constructor.
  apply LE_M1M1'.

  constructor.
  constructor.
  inversion LE_M1M1' as [ _Γ _Γ' _Σ _Σ' LE_LSLS' LE_LHLH' LE_MSMS' LE_MHMH' ]. subst.
  apply LE_LSLS'.

(* ******************************************************************************** *)
(* App *)
rename H into IH1.
rename H0 into IH2.
rename H1 into IH3.

unfold NI_LEval_Prop.
intros Γ' Σ1' Σ4' _lv2' _t LE_M1M1' _EV'. subst. 
inversion _EV' as [ | | | | | | | | | ctx _Γ' _Σ1' Σ2' Σ3' _Σ4' ef' e' l' F' lv1' lv2' EV1' EV2' l0' | | | | | ]. subst.

assert (LE_M2M2'_FF' : LowEq_Memory (Γ, Σ2) (Γ', Σ2')/\ LowEq_LValue (LVFun l F) (LVFun l' F')).
eapply IH1.
  trivial.
  apply LE_M1M1'.
  apply EV1'.

assert (LE_M3M3'_lv1lv1' : LowEq_Memory (Γ, Σ3) (Γ', Σ3') /\ LowEq_LValue lv1 lv1').
eapply IH2.
  trivial.
  apply LE_M2M2'_FF'.
  apply EV2'.

destruct LE_M2M2'_FF' as [LE_M2M2' LE_LVFF'].
inversion LE_LVFF' as [ | _F _F' LE_FF' | | | | _v1 _v2 LO LO' ].

(* l = l' = L *)
subst.

assert (LE_M4M4'_lv2lv2' : LowEq_Memory (Γ, Σ4) (Γ', Σ4') /\ LowEq_LValue lv2 lv2').
eapply IH3.
  trivial.
  apply LE_M3M3'_lv1lv1'.
  apply LE_FF'.
  apply LE_M3M3'_lv1lv1'.
  apply l0'.

rewrite UpgradePublic.
rewrite UpgradePublic.
apply LE_M4M4'_lv2lv2'.

(* l = l' = H *)
unfold LabelOf in LO, LO'.
subst.

assert (LE_M3M4 : LowEq_Memory (Γ, Σ3) (Γ, Σ4)).
eapply Conf_LFunEval.
  apply l2.
  auto.
  eapply LowEq_Memory_Refl.
  apply LE_M3M3'_lv1lv1'.

assert (LE_M3'M4' : LowEq_Memory (Γ', Σ3') (Γ', Σ4')).
eapply Conf_LFunEval.
  apply l0'.
  auto.
  eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M3M3'_lv1lv1'.

assert (LE_M4M4' : LowEq_Memory (Γ, Σ4) (Γ', Σ4')).
eapply LowEq_Memory_Trans.
  apply LowEq_Memory_Symm.
  apply LE_M3M4.
  eapply LowEq_Memory_Trans.
  apply LE_M3M3'_lv1lv1'.
  apply LE_M3'M4'.

constructor.
  apply LE_M4M4'.
  apply LowEq_Secret.

(* ******************************************************************************** *)
(* Lib *)
unfold NI_LEval_Prop.
intros Γ' Σ1' _Σ1' lv2' _t LE_M1M1' _EV'. subst. 
inversion _EV' as [ | | | | | | | | | | ctx _Γ' Σ' f' uv' γ' lv' l' l0' r' | | | | ]. subst _Σ1' Σ1' lv2'.  subst.

assert (exLE_uvuv' :  exists l, LowEq_UValue l uv uv').
eapply LabelabilityOfUnlabeledExecution.
  apply LE_M1M1'.
  apply uevVar with (x := f13). 
  apply l.
  apply uevVar with (x:= f13).
  apply l'.
  
destruct exLE_uvuv' as [ul LE_uvuv'].

assert (EQ : Some (MVModel γ) = Some (MVModel γ')).
eapply NI_LookupM.
  apply LE_M1M1'.
  apply l0.
  apply l0'.

injection EQ. intro. subst γ'. clear EQ.

assert (LE_lvlv' : LowEq_LValue lv lv').
eapply UnlabeledCorrectnessOfLibraryModels.
  apply LE_M1M1'.
  apply LE_uvuv'.
  apply r.
  apply r'.

constructor. 
  apply LE_M1M1'.
  apply LE_lvlv'.

(* ******************************************************************************** *)
(* Nil *)
unfold NI_LEval_Prop.
intros Γ' Σ1' Σ2' lv' _t LE_M1M1' _EV'.
inversion _EV'. subst.
constructor.
  apply LE_M1M1'.
  constructor.

(* ******************************************************************************** *)
(* Cons *)

rename H into IH1.
rename H0 into IH2.

unfold NI_LEval_Prop.
intros Γ' Σ1' Σ3' _v2' _t LE_M1M1' _EV'. subst.
inversion _EV' as [ | | | | | | | | | | |  | ctx _Γ' _Σ1' Σ2' _Σ3' e1' e2' v1' v2' EV1' EV2' | | ].
subst.

assert (LE_M2M2'_v1v1' : LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\ LowEq_LValue v1 v1').
eapply IH1.
  trivial.
  apply LE_M1M1'.
  apply EV1'.

assert (LE_M3M3'_v2v2' : LowEq_Memory (Γ, Σ3) (Γ', Σ3') /\ LowEq_LValue v2 v2').
eapply IH2.
  trivial.
  apply LE_M2M2'_v1v1'.
  apply EV2'.

split.
  apply LE_M3M3'_v2v2'.
  constructor.
  constructor.
  apply LE_M2M2'_v1v1'.
  constructor.
  apply LE_M3M3'_v2v2'.

(* ******************************************************************************** *)
(* Head *)
rename H into IH.
unfold NI_LEval_Prop.
intros Γ' Σ1' Σ3' _lv' _t LE_M1M1' _EV'. subst.
inversion _EV' as [ | | | | | | | | | | | | | ctx' _Γ' _Σ1' Σ2' _Σ3' LHd' LTl' l' e' lv' l0' l1' | ]. subst.

assert (LE_M2M2'_lvlv' : LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\ LowEq_LValue (LVCons l LHd LTl) (LVCons l' LHd' LTl')).
eapply IH.
  trivial.
  apply LE_M1M1'.
  apply l0'.

destruct LE_M2M2'_lvlv' as [LE_M2M2' LE_LL'].
inversion LE_LL' as [ | | | | _LHd _LHd' _LTl _LTl' LE_HdHd' LE_TlTl' | v v' EQ1 EQ2 ]; subst.

(* l = l' = L *)
assert (LE_M3M3'_lvlv' : LowEq_Memory (Γ, Σ3) (Γ', Σ3') /\ LowEq_LValue lv lv').
eapply NI_LHeadEval.
  apply l1.
  trivial.
  apply LE_M2M2'.
  apply LE_HdHd'.
  apply l1'.

rewrite UpgradePublic, UpgradePublic.
apply LE_M3M3'_lvlv'.

(* l = l' = H *)
unfold LabelOf in EQ1, EQ2. subst.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply Conf_LHeadEval.
  apply l1.
  trivial.
  eapply LowEq_Memory_Refl.
  apply LE_M2M2'.
 
assert (LE_M2'M3' : LowEq_Memory (Γ', Σ2') (Γ', Σ3')).
eapply Conf_LHeadEval.
  apply l1'.
  trivial.
  eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M2M2'.

split.

eapply LowEq_Memory_Trans.
  apply LowEq_Memory_Symm.
  apply LE_M2M3.
  eapply LowEq_Memory_Trans.
  apply LE_M2M2'.
  apply LE_M2'M3'.

apply LowEq_Secret.

(* ******************************************************************************** *)
(* Tail *)
rename H into IH.
unfold NI_LEval_Prop.
intros Γ' Σ1' Σ3' _lv' _t LE_M1M1' _EV'. subst.
inversion _EV' as [ | | | | | | | | | | | | | | ctx' _Γ' _Σ1' Σ2' _Σ3' LHd' LTl' l' e' lv' l0' l1' ]. subst.

assert (LE_M2M2'_lvlv' : LowEq_Memory (Γ, Σ2) (Γ', Σ2') /\ LowEq_LValue (LVCons l LHd LTl) (LVCons l' LHd' LTl')).
eapply IH.
  trivial.
  apply LE_M1M1'.
  apply l0'.

destruct LE_M2M2'_lvlv' as [LE_M2M2' LE_LL'].
inversion LE_LL' as [ | | | | _LHd _LHd' _LTl _LTl' LE_HdHd' LE_TlTl' | v v' EQ1 EQ2 ]; subst.

(* l = l' = L *)
assert (LE_M3M3'_lvlv' : LowEq_Memory (Γ, Σ3) (Γ', Σ3') /\ LowEq_LValue lv lv').
eapply NI_LTailEval.
  apply l1.
  trivial.
  apply LE_M2M2'.
  apply LE_TlTl'.
  apply l1'.

rewrite UpgradePublic, UpgradePublic.
apply LE_M3M3'_lvlv'.

(* l = l' = H *)
unfold LabelOf in EQ1, EQ2. subst.

assert (LE_M2M3 : LowEq_Memory (Γ, Σ2) (Γ, Σ3)).
eapply Conf_LTailEval.
  apply l1.
  trivial.
  eapply LowEq_Memory_Refl.
  apply LE_M2M2'.

assert (LE_M2'M3' : LowEq_Memory (Γ', Σ2') (Γ', Σ3')).
eapply Conf_LTailEval.
  apply l1'.
  trivial.
  eapply LowEq_Memory_Refl.
  apply LowEq_Memory_Symm.
  apply LE_M2M2'.

split.

eapply LowEq_Memory_Trans.
  apply LowEq_Memory_Symm.
  apply LE_M2M3.
  eapply LowEq_Memory_Trans.
  apply LE_M2M2'.
  apply LE_M2'M3'.

apply LowEq_Secret.

(* ******************************************************************************** *)
(* For UEval there is nothing to prove *)

unfold NI_UEval_Prop; trivial.
unfold NI_UEval_Prop; trivial.
unfold NI_UEval_Prop; trivial.
unfold NI_UEval_Prop; trivial.
unfold NI_UEval_Prop; trivial.
unfold NI_UEval_Prop; trivial.
unfold NI_UEval_Prop; trivial.
unfold NI_UEval_Prop; trivial.
unfold NI_UEval_Prop; trivial.
unfold NI_UEval_Prop; trivial.
unfold NI_UEval_Prop; trivial.
unfold NI_UEval_Prop; trivial.
unfold NI_UEval_Prop; trivial.
unfold NI_UEval_Prop; trivial.
Qed.
